# -*- coding: utf-8 -*-
"""
Created on 2020-05-11

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict
from rescupy.utils import set_units as utils_set_units

class Cell:
   """
   cell class.
   
   Attributes
   ----------
   avec : 2D array
      Domain vectors.
   boundary : 1D array
      Boundary type for each face of the parallelepipedic domain.
   bravais : string
      Bravais lattice.
   grid : 1D array
      Number of grid points along each dimension.
   cres : float
      Grid resolution

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):
      # default values
      defDict = {
         "avec"      : [0. for i in range(9)],
         "avecUnits" : "ang",
         "boundary"  : [0, 0, 0, 0, 0, 0],
         "bravais"   : None,
         "grid"      : None,
         "lres"      : None,
         "lresUnits" : "ang",
      }
      # required status
      reqDict = {
         "avec"     : True,
         "avecUnits": False,
         "boundary" : False,
         "bravais"  : False,
         "grid"     : False,
         "lres"     : False,
         "lresUnits": False,
      }
      
      init_from_dict(self, reqDict, defDict, inpDict)

      self.avec = np.array(self.avec).reshape((3,3))

      if self.bravais is None:
         self.bravais = self.get_bravais_lattice()
         
      if self.grid is None: 
         try: 
            if self.lresUnits != self.avecUnits:
               raise NameError('Inconsistent units between cell.avec and cell.lres.')
            lres = self.lres
            clen = self.lengths()
            self.grid = np.ceil(clen / lres).astype(int)
         except ValueError:
            raise NameError('Input file is missing a value for parameter cell.grid.')

      if self.lres is None: 
         try: 
            grid = inpDict['grid']
            clen = self.lengths()
            self.lres = np.mean(clen / grid)
            self.lresUnits = self.avecUnits
         except ValueError:
            raise NameError('Input file is missing a value for parameter cell.lres.')

      if isinstance(self.grid, list):
         self.grid = np.array(self.grid)

   def asdict(self):
      return utils_asdict(self)

   def set_units(self, units):
      return utils_set_units(self, units)

   def get_dx(self, axis):
      """
      Returns the grid node-to-node distance along a given direction.

      Parameters:
         axis (int): Lattice vector index.

      Returns: 
         dx (float): Node-to-node distance
      """
      return self.get_length(axis) / self.grid[axis]

   def get_grid(self, axis):
      """
      Returns the coordinates of the grid nodes along a given direction.

      Parameters:
         axis (int): Axis along which the grid node coordinates are sought.

      Returns: 
         x (ndarray): Numpy array containing the grid node coordinates.
      """
      n = self.grid[axis]
      u = np.zeros((n,3))
      u[:,axis] = np.array([i for i in range(1,n+1)]) / n
      x = np.matmul(u,self.avec)
      return x

   def get_length(self, axis):
      """
      Returns the length of a lattice vector.

      Parameters:
         axis (int): Lattice vector index.

      Returns: 
         l (float): Length of the lattice vector.
      """
      return np.linalg.norm(self.avec[axis,:])

   def get_lengths(self):
      """
      Returns the lengths of the lattice vectors.

      Parameters:

      Returns: 
         l (ndarray): Lengths of the lattice vectors.
      """
      return np.linalg.norm(self.avec, axis=1)

   def get_bravais_lattice(self, tol = 1e-3):
      return get3DCrystalStructure(self.avec, tol = tol)

   def get_standard_lattice(self, tol = 1e-3):
      if self.bravais is None:
         self.bravais = self.get_bravais_lattice(tol = tol)
      return getStandardPrimitiveCellVectors(self.bravais, self.avec, tol = tol)

   def lengths(self):
      avec = self.avec
      return np.linalg.norm(avec, axis=1)

   def reciprocal(self):
      avec = self.avec
      bvec = 2 * np.pi * np.linalg.inv(avec).T
      if self.avecUnits == "ang":
         bvecUnits = "invang"
      elif self.avecUnits == "bohr":
         bvecUnits = "invbohr"
      return bvec, bvecUnits

def get3DCrystalStructure(primitiveCellVectors, tol = 1e-3):
   """
   Return the crystal structure of given three linearly independent lattice vectors.

   Parameters
   ----------
   primitiveCellVectors: 2D array
      Lattice vectors (column-vectors).
      
   tol: scalar
      Tolerance.

   Returns
   ----------
   crystalStructure: string
      Crystal structure.
      
   """
   #
   # 14 cases:
   #     case 01: {'Cubic','CUB','SimpleCubic','SC','cP'})
   #     case 02: {'FaceCenteredCubic','FCC','cF'})
   #     case 03: {'BodyCenteredCubic','BCC','cI'})
   #     case 04: {'Tetragonal','TET','tP'})
   #     case 05: {'BodyCenteredTetragonal','BCT','tI'})
   #     case 06: {'Orthorhombic','ORC','oP'})
   #     case 07: {'FaceCenteredOrthorhombic','ORCF','oF'})
   #     case 08: {'BodyCenteredOrthorhombic','ORCI','oI'})
   #     case 09: {'CCenteredOrthorhombic','ORCC','oS'})
   #     case 10: {'Hexagonal','HEX','HCP','hP'})
   #     case 11: {'Rhombohedral','RHL','hR'})
   #     case 12: {'Monoclinic','MCL','mP'})
   #     case 13: {'CCenteredMonoclinic','MCLC','mS'})
   #     case 14: {'Triclinic','TRI','aP'})
   #
   v1 = primitiveCellVectors[0,:]
   v2 = primitiveCellVectors[1,:]
   v3 = primitiveCellVectors[2,:]
   a1 = np.linalg.norm(v1)
   a2 = np.linalg.norm(v2)
   a3 = np.linalg.norm(v3)
   a12 = np.dot(v1,v2)/a1/a2
   a23 = np.dot(v3,v2)/a2/a3
   a31 = np.dot(v1,v3)/a3/a1
   sqrt2 = np.sqrt(2.)
   sqrt3 = np.sqrt(3.)

   # SC
   if abs(a1/a2-1) < tol and abs(a2/a3-1) < tol and abs(a3/a1-1) < tol and \
         abs(a12-0) < tol and abs(a23-0) < tol and abs(a31-0) < tol:
      crystalStructure = 'CUB'
      return crystalStructure

   # FCC
   if abs(a1/a2-1) < tol and abs(a2/a3-1) < tol and abs(a3/a1-1) < tol and \
         abs(a12-1/2) < tol and abs(a23-1/2) < tol and abs(a31-1/2) < tol:
      crystalStructure = 'FCC'
      return crystalStructure

   # FCC-1
   if abs(a2/a3-1) < tol and abs(a3/a1-sqrt2/2) < tol and abs(a1/a2-sqrt2) < tol and \
         abs(2*a23-sqrt2*a31-sqrt2*a12+1) < tol and abs(sqrt2*a12-1) < tol and abs(sqrt2*a31-1) < tol:
      crystalStructure = 'FCC'
      return crystalStructure

   # FCC-2
   if abs(a3/a1-1) < tol and abs(a1/a2-sqrt2/2) < tol and abs(a2/a3-sqrt2) < tol and \
         abs(2*a31-sqrt2*a12-sqrt2*a23+1) < tol and abs(sqrt2*a23-1) < tol and abs(sqrt2*a12-1) < tol:
      crystalStructure = 'FCC'
      return crystalStructure

   # FCC-3
   if abs(a1/a2-1) < tol and abs(a2/a3-sqrt2/2) < tol and abs(a3/a1-sqrt2) < tol and \
         abs(2*a12-sqrt2*a23-sqrt2*a31+1) < tol and abs(sqrt2*a31-1) < tol and abs(sqrt2*a23-1) < tol:
      crystalStructure = 'FCC'
      return crystalStructure

   # BCC
   if abs(a1/a2-1) < tol and abs(a2/a3-1) < tol and abs(a3/a1-1) < tol and \
         abs(a12+1/3) < tol and abs(a23+1/3) < tol and abs(a31+1/3) < tol:
      crystalStructure = 'BCC'
      return crystalStructure

   # BCC-3
   if abs(a1/a2-1) < tol and abs(a2/a3-2/sqrt3) < tol and abs(a3/a1-sqrt3/2) < tol and \
         abs(a12-0) < tol and abs(a23-1/sqrt3) < tol and abs(a31-1/sqrt3) < tol:
      crystalStructure = 'BCC'
      return crystalStructure

   # BCC-1
   if abs(a2/a3-1) < tol and abs(a3/a1-2/sqrt3) < tol and abs(a1/a2-sqrt3/2) < tol and \
         abs(a23-0) < tol and abs(a31-1/sqrt3) < tol and abs(a12-1/sqrt3) < tol:
      crystalStructure = 'BCC'
      return crystalStructure

   # BCC-2
   if abs(a3/a1-1) < tol and abs(a1/a2-2/sqrt3) < tol and abs(a2/a3-sqrt3/2) < tol and \
         abs(a31-0) < tol and abs(a12-1/sqrt3) < tol and abs(a23-1/sqrt3) < tol:
      crystalStructure = 'BCC'
      return crystalStructure

   # Tetragonal
   if (abs(a1/a2-1) < tol or abs(a2/a3-1) < tol or abs(a3/a1-1) < tol) and \
         abs(a12-0) < tol and abs(a23-0) < tol and abs(a31-0) < tol:
      crystalStructure = 'TET'
      return crystalStructure

   # BodyCenteredTetragonal
   if abs(a1/a2-1) < tol and abs(a2/a3-1) < tol and abs(a3/a1-1) < tol and \
         (abs(a1/a2-a2/a1+a31/a2*a3-a23/a1*a3) < tol or abs(a2/a3-a3/a2+a12/a3*a1-a31/a2*a1) < tol  or abs(a3/a1-a1/a3+a23/a1*a2-a12/a3*a2) < tol) and \
         abs(a12+a23+a31+1) < tol:
      crystalStructure = 'BCT'
      return crystalStructure

   # BodyCenteredTetragonal-3
   if abs(a1/a2-1) < tol and \
         abs(a12-0) < tol and abs(2*a23-a2/a3) < tol and abs(2*a31-a1/a3) < tol:
      crystalStructure = 'BCT'
      return crystalStructure

   # BodyCenteredTetragonal-1
   if abs(a2/a3-1) < tol and \
         abs(a23-0) < tol and abs(2*a31-a3/a1) < tol and abs(2*a12-a2/a1) < tol:
      crystalStructure = 'BCT'
      return crystalStructure

   # BodyCenteredTetragonal-2
   if abs(a3/a1-1) < tol and \
         abs(a31-0) < tol and abs(2*a12-a1/a2) < tol and abs(2*a23-a3/a2) < tol:
      crystalStructure = 'BCT'
      return crystalStructure

   # Orthorhombic
   if abs(a12-0) < tol and abs(a23-0) < tol and abs(a31-0) < tol:
      crystalStructure = 'ORC'
      return crystalStructure

   # FaceCenteredOrthorhombic
   if abs(2*a12-a2/a1-a1/a2+a3/a2*a3/a1) < tol and abs(2*a23-a3/a2-a2/a3+a1/a3*a1/a2) < tol and abs(2*a31-a1/a3-a3/a1+a2/a1*a2/a3) < tol:
      crystalStructure = 'ORCF'
      return crystalStructure

   # FaceCenteredOrthorhombic-1
   if abs(4*a23*a2/a1*a3/a1-2*a31*a3/a1-2*a12*a2/a1+1) < tol and abs(2*a12*a2/a1-1) < tol and abs(2*a31*a3/a1-1) < tol:
      crystalStructure = 'ORCF'
      return crystalStructure

   # FaceCenteredOrthorhombic-2
   if abs(4*a31*a3/a2*a1/a2-2*a12*a1/a2-2*a23*a3/a2+1) < tol and abs(2*a23*a3/a2-1) < tol and abs(2*a12*a1/a2-1) < tol:
      crystalStructure = 'ORCF'
      return crystalStructure

   # FaceCenteredOrthorhombic-3
   if abs(4*a12*a1/a3*a2/a3-2*a23*a2/a3-2*a31*a1/a3+1) < tol and abs(2*a31*a1/a3-1) < tol and abs(2*a23*a2/a3-1) < tol:
      crystalStructure = 'ORCF'
      return crystalStructure

   # BodyCenteredOrthorhombic
   if abs(a1/a2-1) < tol and abs(a2/a3-1) < tol and abs(a3/a1-1) < tol and \
         abs(a12+a23+a31+1) < tol:
      crystalStructure = 'ORCI'
      return crystalStructure

   # BodyCenteredOrthorhombic-3
   if abs(a12-0) < tol and abs(2*a23-a2/a3) < tol and abs(2*a31-a1/a3) < tol:
      crystalStructure = 'ORCI'
      return crystalStructure

   # BodyCenteredOrthorhombic-1
   if abs(a23-0) < tol and abs(2*a31-a3/a1) < tol and abs(2*a12-a2/a1) < tol:
      crystalStructure = 'ORCI'
      return crystalStructure

   # BodyCenteredOrthorhombic-2
   if abs(a31-0) < tol and abs(2*a12-a1/a2) < tol and abs(2*a23-a3/a2) < tol:
      crystalStructure = 'ORCI'
      return crystalStructure

   # CCenteredOrthorhombic: SideCenteredOrthorhombic-3
   if abs(a1/a2-1) < tol and \
         abs(a23-0) < tol and abs(a31-0) < tol and \
         abs(abs(a12)-1/2) > tol:
      crystalStructure = 'ORCC'
      return crystalStructure

   # CCenteredOrthorhombic: SideCenteredOrthorhombic-1
   if abs(a2/a3-1) < tol and \
         abs(a31-0) < tol and abs(a12-0) < tol and \
         abs(abs(a23)-1/2) > tol:
      crystalStructure = 'ORCC'
      return crystalStructure

   # CCenteredOrthorhombic: SideCenteredOrthorhombic-2
   if abs(a3/a1-1) < tol and \
         abs(a12-0) < tol and abs(a23-0) < tol and \
         abs(abs(a31)-1/2) > tol:
      crystalStructure = 'ORCC'
      return crystalStructure

   # Hexagonal
   if abs(a1/a2-1) < tol and \
         abs(abs(a12)-1/2) < tol and abs(a23-0) < tol and abs(a31-0) < tol:
      crystalStructure = 'HEX'
      return crystalStructure
   elif abs(a2/a3-1) < tol and \
         abs(abs(a23)-1/2) < tol and abs(a31-0) < tol and abs(a12-0) < tol:
      crystalStructure = 'HEX'
      return crystalStructure
   elif abs(a3/a1-1) < tol and \
         abs(abs(a31)-1/2) < tol and abs(a12-0) < tol and abs(a23-0) < tol:
      crystalStructure = 'HEX'
      return crystalStructure

   # Rhombohedral
   if abs(a1/a2-1) < tol and abs(a2/a3-1) < tol and abs(a3/a1-1) < tol and \
         abs(a12/a23-1) < tol and abs(a23/a31-1) < tol and abs(a31/a12-1) < tol:
      crystalStructure = 'RHL'
      return crystalStructure

   # Monoclinic
   if np.count_nonzero(abs(np.array([a12, a23, a31])-0.) < tol) == 2:
      crystalStructure = 'MCL'
      return crystalStructure

   # CCenteredMonoclinic: SideCenteredMonoclinic-3
   if abs(a1/a2-1) < tol and \
         abs(abs(a23)-abs(a31)) < tol:
      crystalStructure = 'MCLC'
      return crystalStructure

   # CCenteredMonoclinic: SideCenteredMonoclinic-1
   if abs(a2/a3-1) < tol and \
         abs(abs(a31)-abs(a12)) < tol:
      crystalStructure = 'MCLC'
      return crystalStructure

   # CCenteredMonoclinic: SideCenteredMonoclinic-2
   if abs(a3/a1-1) < tol and \
         abs(abs(a12)-abs(a23)) < tol:
      crystalStructure = 'MCLC'
      return crystalStructure

   # Triclinic
   crystalStructure = 'TRI'
   return crystalStructure

import numpy as np

def getStandardPrimitiveCellVectors(crystalStructure,primitiveCellVectors,tol=1e-3):
   """
   For 0D or 1D system,there is only 1, case.
   
   5 cases for 2D system:
       case 1: {'Square2D','SQU2D'}
       case 2: {'Rectangular2D','Rectangle2D','REC2D'}
       case 3: {'Hexagonal2D','Hexagon2D','HEX2D'}
       case 4: {'Rhombic2D','Rhombus2D','CenteredRectangular2D','RHO2D'}
       case 5: {'Oblique2D','OBL2D'}
   
   14 cases for 3D system:
       case 01: {'Cubic','CUB','SimpleCubic','SC','cP'})
       case 02: {'FaceCenteredCubic','FCC','cF'})
       case 03: {'BodyCenteredCubic','BCC','cI'})
       case 04: {'Tetragonal','TET','tP'})
       case 05: {'BodyCenteredTetragonal','BCT','tI'})
       case 06: {'Orthorhombic','ORC','oP'})
       case 07: {'FaceCenteredOrthorhombic','ORCF','oF'})
       case 08: {'BodyCenteredOrthorhombic','ORCI','oI'})
       case 09: {'CCenteredOrthorhombic','ORCC','oS'})
       case 10: {'Hexagonal','HEX','HCP','hP'})
       case 11: {'Rhombohedral','RHL','hR'})
       case 12: {'Monoclinic','MCL','mP'})
       case 13: {'CCenteredMonoclinic','MCLC','mS'})
       case 14: {'Triclinic','TRI','aP'})
   """
   primitiveCellVectors = primitiveCellVectors.transpose()
   v1 = primitiveCellVectors[:,0]
   v2 = primitiveCellVectors[:,1]
   v3 = primitiveCellVectors[:,2]
   a1 = np.linalg.norm(v1)
   a2 = np.linalg.norm(v2)
   a3 = np.linalg.norm(v3)
   a12 = np.dot(v1,v2)/a1/a2
   a23 = np.dot(v3,v2)/a2/a3
   a31 = np.dot(v1,v3)/a3/a1
   sqrt2 = np.sqrt(2.)
   sqrt3 = np.sqrt(3.)

   if crystalStructure == '0D':
      # 0D: isolate :: [10, 10, 10]
      spcv = primitiveCellVectors
   elif crystalStructure == '1D':
      # 1D: along z direction. c << a,b :: [10, 10, 1]
      # pcv = primitiveCellVectors
      I = np.argsort(-[a1, a2, a3]) # descending
      spcv = primitiveCellVectors[:,I]
   elif crystalStructure in ['Square2D','SQU2D']:
      # 2D1: along x-y plane. c >> a,b
      #     [[1, 0, 0]',[0,1,0]',[0, 0, 10]'] :: [1, 1, 10]
      #     analog: Tetragonal
      I = np.argsort([a1, a2, a3])
      spcv = primitiveCellVectors[:,I]
   elif crystalStructure in ['Rectangular2D','Rectangle2D','REC2D']:
      # 2D2: along x-y plane. c >> a,b
      #     [[1, 0, 0]',[0,2,0]',[0, 0, 10]'] :: [1, 1, 10]
      #     analog: Orthorhombic
      I = np.argsort([a1, a2, a3])
      spcv = primitiveCellVectors[:,I]
   elif crystalStructure in ['Hexagonal2D','Hexagon2D','HEX2D']:
      # 2D3: along x-y plane. c >> a,b
      #     [[1/2, -sqrt3/2 0]',[1/2,sqrt3/2,0]',[0, 0, 10]'] :: [1, 1, 10]
      #     analog: Hexagonal
      I = np.argsort([a1, a2, a3])
      spcv = primitiveCellVectors[:,I]
      if np.linalg.norm(spcv[:,0]+spcv[:,1]) > np.linalg.norm(spcv[:,0]-spcv[:,1]):
         spcv[:,1] = -spcv[:,1]
   elif crystalStructure in ['Rhombic2D','Rhombus2D','CenteredRectangular2D','RHO2D']:
      # 2D4:  along x-y plane. c >> a,b
      #     [[1/2, -2/2 0]',[1/2,2/2,0]',[0, 0, 10]'] :: [1, 2 10]
      #     analog: CCenteredOrthorhombic
      I = np.argsort([a1, a2, a3])
      spcv = primitiveCellVectors[:,I]
      if np.linalg.norm(spcv[:,0]+spcv[:,1]) > np.linalg.norm(spcv[:,0]-spcv[:,1]):
         spcv[:,1] = -spcv[:,1]
   elif crystalStructure in ['Oblique2D','OBL2D']:
      # 2D5:  along x-y plane. c >> a,b
      #     [[2 0, 0]',[3*cos(pi/5),3*sin(pi/5),0]',[0, 0, 10]'] :: [2 3 10]
      #     analog: 'Monoclinic'
      I = np.argsort([a1, a2, a3])
      spcv = primitiveCellVectors[:,I]
      if dot(spcv[:,0],spcv[:,1]) < 0:
         spcv[:,0] = -spcv[:,0]
   elif crystalStructure in ['Cubic','CUB','SimpleCubic','SC','cP']:
      # 1: ucv = pcv = [[1, 0, 0]',[0, 1, 0]',[0, 0, 1]']
      ucv = primitiveCellVectors
      sucv = getStandardUnitCellVectors(crystalStructure,ucv,tol)
      spcv = sucv
   elif crystalStructure in ['FaceCenteredCubic','FCC','cF']:
      # 2: pcv = [[0, 1/2, 1/2]',[1/2,0,1/2]',[1/2, 1/2, 0]']
      #    ucv = [[1, 0, 0]',[0, 1, 0]',[0, 0, 1]']
      pcv = primitiveCellVectors
      if abs(a1/a2-1) < tol and abs(a2/a3-1) < tol and abs(a3/a1-1) < tol and \
            abs(a12-1/2) < tol and abs(a23-1/2) < tol and abs(a31-1/2) < tol:
         ucv = np.linalg.inv(np.array([[0, 1/2, 1/2],[1/2,0,1/2],[1/2, 1/2, 0]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(a2/a3-1) < tol and abs(a3/a1-sqrt2/2) < tol and abs(a1/a2-sqrt2) < tol and \
            abs(2*a23-sqrt2*a31-sqrt2*a12+1) < tol and abs(sqrt2*a12-1) < tol and abs(sqrt2*a31-1) < tol:
         ucv = np.linalg.inv(np.array([[1, 0, 0],[1/2,0,1/2],[1/2, 1/2, 0]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(a3/a1-1) < tol and abs(a1/a2-sqrt2/2) < tol and abs(a2/a3-sqrt2) < tol and \
            abs(2*a31-sqrt2*a12-sqrt2*a23+1) < tol and abs(sqrt2*a23-1) < tol and abs(sqrt2*a12-1) < tol:
         ucv = np.linalg.inv(np.array([[0, 1/2, 1/2],[0, 1, 0],[1/2, 1/2, 0]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(a1/a2-1) < tol and abs(a2/a3-sqrt2/2) < tol and abs(a3/a1-sqrt2) < tol and \
            abs(2*a12-sqrt2*a23-sqrt2*a31+1) < tol and abs(sqrt2*a31-1) < tol and abs(sqrt2*a23-1) < tol:
         ucv = np.linalg.inv(np.array([[0, 1/2, 1/2],[1/2,0,1/2],[0, 0, 1]]).T)
         ucv = np.matmul(pcv, ucv)
      sucv = getStandardUnitCellVectors(crystalStructure,ucv,tol)
      spcv = np.matmul(sucv, np.array([[0, 1/2, 1/2],[1/2,0,1/2],[1/2, 1/2, 0]]).T)
   elif crystalStructure in ['BodyCenteredCubic','BCC','cI']:
      # 3: pcv = [[-1/2, 1/2, 1/2]',[1/2,-1/2,1/2]',[1/2, 1/2, -1/2]']
      #    ucv = [[1, 0, 0]',[0, 1, 0]',[0, 0, 1]']
      pcv = primitiveCellVectors
      if abs(a1/a2-1) < tol and abs(a2/a3-1) < tol and abs(a3/a1-1) < tol and \
            abs(a12+1/3) < tol and abs(a23+1/3) < tol and abs(a31+1/3) < tol:
         ucv = np.linalg.inv(np.array([[-1/2, 1/2, 1/2],[1/2,-1/2,1/2],[1/2, 1/2, -1/2]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(a1/a2-1) < tol and abs(a2/a3-2/sqrt3) < tol and abs(a3/a1-sqrt3/2) < tol and \
            abs(a12-0) < tol and abs(a23-1/sqrt3) < tol and abs(a31-1/sqrt3) < tol:
         ucv = np.linalg.inv(np.array([[1, 0, 0],[0, 1, 0],[1/2, 1/2, 1/2]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(a2/a3-1) < tol and abs(a3/a1-2/sqrt3) < tol and abs(a1/a2-sqrt3/2) < tol and \
            abs(a23-0) < tol and abs(a31-1/sqrt3) < tol and abs(a12-1/sqrt3) < tol:
         ucv = np.linalg.inv(np.array([[1/2, 1/2, 1/2],[0, 1, 0],[0, 0, 1]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(a3/a1-1) < tol and abs(a1/a2-2/sqrt3) < tol and abs(a2/a3-sqrt3/2) < tol and \
            abs(a31-0) < tol and abs(a12-1/sqrt3) < tol and abs(a23-1/sqrt3) < tol:
         ucv = np.linalg.inv(np.array([[1, 0, 0],[1/2, 1/2, 1/2],[0, 0, 1]]).T)
         ucv = np.matmul(pcv, ucv)
      sucv = getStandardUnitCellVectors(crystalStructure,ucv,tol)
      spcv = np.matmul(sucv, np.array([[-1/2, 1/2, 1/2],[1/2,-1/2,1/2],[1/2, 1/2, -1/2]]).T)
   elif crystalStructure in ['Tetragonal','TET','tP']:
      # 4: ucv = pcv = [[1, 0, 0]',[0, 1, 0]',[0, 0, 5]']
      ucv = primitiveCellVectors
      sucv = getStandardUnitCellVectors(crystalStructure,ucv,tol)
      spcv = sucv
   elif crystalStructure in ['BodyCenteredTetragonal','BCT','tI']:
      # 5: pcv = [[-1/2, 1/2, 5/2]',[1/2,-1/2,5/2]',[1/2, 1/2, -5/2]']
      #    ucv = [[1, 0, 0]',[0, 1, 0]',[0, 0, 5]']
      pcv = primitiveCellVectors
      if abs(a1/a2-1) < tol and abs(a2/a3-1) < tol and abs(a3/a1-1) < tol and \
            (abs(a1/a2-a2/a1+a31/a2*a3-a23/a1*a3) < tol or abs(a2/a3-a3/a2+a12/a3*a1-a31/a2*a1) < tol or \
               abs(a3/a1-a1/a3+a23/a1*a2-a12/a3*a2) < tol) and abs(a12+a23+a31+1) < tol:
         ucv = np.linalg.inv(np.array([[-1/2, 1/2, 1/2],[1/2,-1/2,1/2],[1/2, 1/2, -1/2]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(a1/a2-1) < tol and \
            abs(a12-0) < tol and abs(2*a23-a2/a3) < tol and abs(2*a31-a1/a3) < tol:
         ucv = np.linalg.inv(np.array([[1, 0, 0],[0, 1, 0],[1/2, 1/2, 1/2]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(a2/a3-1) < tol and \
            abs(a23-0) < tol and abs(2*a31-a3/a1) < tol and abs(2*a12-a2/a1) < tol:
         ucv = np.linalg.inv(np.array([[1/2, 1/2, 1/2],[0, 1, 0],[0, 0, 1]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(a3/a1-1) < tol and \
            abs(a31-0) < tol and abs(2*a12-a1/a2) < tol and abs(2*a23-a3/a2) < tol:
         ucv = np.linalg.inv(np.array([[1, 0, 0],[1/2, 1/2, 1/2],[0, 0, 1]]).T)
         ucv = np.matmul(pcv, ucv)
      sucv = getStandardUnitCellVectors(crystalStructure,ucv,tol)
      spcv = np.matmul(sucv, np.array([[-1/2, 1/2, 1/2],[1/2,-1/2,1/2],[1/2, 1/2, -1/2]]).T)
   elif crystalStructure in ['Orthorhombic','ORC','oP']:
      # 6: ucv = pcv = [[1, 0, 0]',[0, 2 0]',[0, 0, 3]']
      ucv = primitiveCellVectors
      sucv = getStandardUnitCellVectors(crystalStructure,ucv,tol)
      spcv = sucv
   elif crystalStructure in ['FaceCenteredOrthorhombic','ORCF','oF']:
      # 7: pcv = [[0, 2/2 3/2]',[1/2,0,3/2]',[1/2, 2/2 0]']
      #    ucv = [[1, 0, 0]',[0, 2 0]',[0, 0, 3]']
      pcv = primitiveCellVectors
      if abs(2*a12-a2/a1-a1/a2+a3/a2*a3/a1) < tol and abs(2*a23-a3/a2-a2/a3+a1/a3*a1/a2) < tol and \
         abs(2*a31-a1/a3-a3/a1+a2/a1*a2/a3) < tol:
         ucv = np.linalg.inv(np.array([[0, 1/2, 1/2],[1/2,0,1/2],[1/2, 1/2, 0]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(4*a23*a2/a1*a3/a1-2*a31*a3/a1-2*a12*a2/a1+1) < tol and abs(2*a12*a2/a1-1) < tol and \
         abs(2*a31*a3/a1-1) < tol:
         ucv = np.linalg.inv(np.array([[1, 0, 0],[1/2,0,1/2],[1/2, 1/2, 0]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(4*a31*a3/a2*a1/a2-2*a12*a1/a2-2*a23*a3/a2+1) < tol and abs(2*a23*a3/a2-1) < tol and \
         abs(2*a12*a1/a2-1) < tol:
         ucv = np.linalg.inv(np.array([[0, 1/2, 1/2],[0, 1, 0],[1/2, 1/2, 0]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(4*a12*a1/a3*a2/a3-2*a23*a2/a3-2*a31*a1/a3+1) < tol and abs(2*a31*a1/a3-1) < tol and \
         abs(2*a23*a2/a3-1) < tol:
         ucv = np.linalg.inv(np.array([[0, 1/2, 1/2],[1/2,0,1/2],[0, 0, 1]]).T)
         ucv = np.matmul(pcv, ucv)
      sucv = getStandardUnitCellVectors(crystalStructure,ucv,tol)
      spcv = np.matmul(sucv, np.array([[0, 1/2, 1/2],[1/2,0,1/2],[1/2, 1/2, 0]]).T)
   elif crystalStructure in ['BodyCenteredOrthorhombic','ORCI','oI']:
      # 8: pcv = [[-1/2, 2/2 3/2]',[1/2,-2/2,3/2]',[1/2, 2/2 -3/2]']
      #    ucv = [[1, 0, 0]',[0, 2 0]',[0, 0, 3]']
      pcv = primitiveCellVectors
      if abs(a1/a2-1) < tol and abs(a2/a3-1) < tol and abs(a3/a1-1) < tol and abs(a12+a23+a31+1) < tol:
         ucv = np.linalg.inv(np.array([[-1/2, 1/2, 1/2],[1/2,-1/2,1/2],[1/2, 1/2, -1/2]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(a12-0) < tol and abs(2*a23-a2/a3) < tol and abs(2*a31-a1/a3) < tol:
         ucv = np.linalg.inv(np.array([[1, 0, 0],[0, 1, 0],[1/2, 1/2, 1/2]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(a23-0) < tol and abs(2*a31-a3/a1) < tol and abs(2*a12-a2/a1) < tol:
         ucv = np.linalg.inv(np.array([[1/2, 1/2, 1/2],[0, 1, 0],[0, 0, 1]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(a31-0) < tol and abs(2*a12-a1/a2) < tol and abs(2*a23-a3/a2) < tol:
         ucv = np.linalg.inv(np.array([[1, 0, 0],[1/2, 1/2, 1/2],[0, 0, 1]]).T)
         ucv = np.matmul(pcv, ucv)
      sucv = getStandardUnitCellVectors(crystalStructure,ucv,tol)
      spcv = np.matmul(sucv, np.array([[-1/2, 1/2, 1/2],[1/2,-1/2,1/2],[1/2, 1/2, -1/2]]).T)
   elif crystalStructure in ['CCenteredOrthorhombic','ORCC','oS']:
      # 9: pcv = [[1/2, -2/2 0]',[1/2,2/2,0]',[0, 0, 3]']
      #    ucv = [[1, 0, 0]',[0, 2 0]',[0, 0, 3]']
      pcv = primitiveCellVectors
      if abs(a1/a2-1) < tol and abs(a23-0) < tol and abs(a31-0) < tol and abs(abs(a12)-1/2) > tol:
         ucv = np.linalg.inv(np.array([[1/2, -1/2, 0],[1/2,1/2,0],[0, 0, 1]]).T)
         ucv = np.matmul(pcv, ucv)
      elif abs(a2/a3-1) < tol and abs(a31-0) < tol and abs(a12-0) < tol and abs(abs(a23)-1/2) > tol:
         ucv = np.linalg.inv(np.array([[1, 0, 0],[0, 1/2, -1/2],[0, 1/2,1/2]]).T)
         ucv = np.matmul(pcv, ucv)
         ucv = ucv[:,[1,2,0]]
      elif abs(a3/a1-1) < tol and abs(a12-0) < tol and abs(a23-0) < tol and abs(abs(a31)-1/2) > tol:
         ucv = np.linalg.inv(np.array([[-1/2, 0, 1/2],[0, 1, 0],[1/2, 0, 1/2]]).T)
         ucv = np.matmul(pcv, ucv)
         ucv = ucv[:,[2,0,1]]
      sucv = getStandardUnitCellVectors(crystalStructure,ucv,tol)
      spcv = np.matmul(sucv, np.array([[1/2, -1/2, 0],[1/2,1/2,0],[0, 0, 1]]).T)
   elif crystalStructure in ['Hexagonal','HEX','HCP','hP']:
      # 10: ucv = pcv = [[1/2, -sqrt3/2 0]',[1/2,sqrt3/2,0]',[0, 0, 3]']
      pcv = primitiveCellVectors
      if abs(1/2-abs(a12)) <= tol:
         spcv = pcv
      elif abs(1/2-abs(a23)) <= tol:
         spcv = pcv[:,[1,2,0]]
      elif abs(1/2-abs(a31)) <= tol:
         spcv = pcv[:,[2,0,1]]
      if np.linalg.norm(spcv[:,0]+spcv[:,1]) > np.linalg.norm(spcv[:,0]-spcv[:,1]):
         spcv[:,1] = -spcv[:,1]
      if np.linalg.det(spcv) < 0:
         spcv[:,2] = -spcv[:,2]
   elif crystalStructure in ['Rhombohedral','RHL','hR']:
      # 11: ucv = pcv = [[cos(pi/5) -sin(pi/5) 0]',[cos(pi/5) sin(pi/5),0]',
      #      [cos(2*pi/5)/cos(pi/5),0,np.sqrt(1-(cos(2*pi/5)/cos(pi/5))^2)]']
      pcv = primitiveCellVectors
      if np.linalg.det(pcv) > 0:
         spcv = pcv
      else:
         spcv = pcv[:,[0,2,1]]
   elif crystalStructure in ['Monoclinic','MCL','mP']:
      # 12: ucv = pcv = [[1, 0, 0]',[0, 2 0]',[0,3*cos(pi/5),3*sin(pi/5)]']
      pcv = primitiveCellVectors
      if abs(a31) <= tol and abs(a12) <= tol:
         spcv = pcv
      elif abs(a12) <= tol and abs(a23) <= tol:
         spcv = pcv[:,[1,2,0]]
      elif abs(a23) <= tol and abs(a31) <= tol:
         spcv = pcv[:,[2,0,1]]
      if np.linalg.norm(spcv[:,1]) > np.linalg.norm(spcv[:,2]):
         spcv[:,[1, 2]] = spcv[:,[2, 1]]
      if np.dot(spcv[:,1],spcv[:,2]) < 0:
         spcv[:,2] = -spcv[:,2]
      if np.linalg.det(spcv) < 0:
         spcv[:,0] = -spcv[:,0]
   elif crystalStructure in ['CCenteredMonoclinic','MCLC','mS']:
      # 13: pcv = [[1/2, 2/2 0]',[-1/2, 2/2 0]',[0,3*cos(pi/5),3*sin(pi/5)]']
      #     ucv = [[1, 0, 0]',[0, 2 0]',[0,3*cos(pi/5),3*sin(pi/5)]']
      pcv = primitiveCellVectors
      if abs(1-a1/a2) <= tol and \
            abs(abs(a23)-abs(a31)) < tol:
         spcv = pcv
      elif abs(1-a2/a3) <= tol and \
            abs(abs(a31)-abs(a12)) <= tol:
         spcv = pcv[:,[1,2,0]]
      elif abs(1-a3/a1) <= tol and \
            abs(abs(a12)-abs(a23)) <= tol:
         spcv = pcv[:,[2,0,1]]
      if np.dot(spcv[:,0],spcv[:,2]) < 0:
         spcv[:,0] = -spcv[:,0]
      if np.dot(spcv[:,1],spcv[:,2]) < 0:
         spcv[:,1] = -spcv[:,1]
      if np.linalg.det(spcv) < 0:
         spcv = spcv[:,[1,0,2]]
   elif crystalStructure in ['Triclinic','TRI','aP']:
      # 14: ucv = pcv = [[1, 0, 0]',[2*cos(pi/3),2*sin(pi/3),0]',[3*cos(pi/4),
      #      3/sin(pi/3)*(cos(pi/5)-cos(pi/4)*cos(pi/3)),
      #      3/sin(pi/3)*np.sqrt(1-(cos(pi/5))^2-(cos(pi/4))^2-(cos(pi/3))^2
      #      + 2*cos(pi/5)*cos(pi/4)*cos(pi/3))]'] :: [1, 2 3],[pi/5,pi/4,pi/3]
      kpcv = np.linalg.inv(primitiveCellVectors).transpose()
      a = np.linalg.norm(kpcv[:,0])
      b = np.linalg.norm(kpcv[:,1])
      c = np.linalg.norm(kpcv[:,2])
      kalpha = np.dot(kpcv[:,1],kpcv[:,2])/b/c
      kbeta = np.dot(kpcv[:,2],kpcv[:,0])/c/a
      kgamma = np.dot(kpcv[:,0],kpcv[:,1])/a/b
      # making all signs of them same
      if kalpha*kbeta > 0:
         if kbeta*kgamma > 0:
            tmp = None
            # +++ or ---
         else:
            # ++- or --+
            kpcv[:,2] = -kpcv[:,2]
      else:
         if kbeta*kgamma > 0:
            # +-- or -++
            kpcv[:,0] = -kpcv[:,0]
         else:
            # +-+ or -+-
            kpcv[:,1] = -kpcv[:,1]
      # making the abs(kgamma) the smallest
      I = np.argsort(-[abs(kalpha),abs(kbeta),abs(kgamma)]) # descending
      kpcv = kpcv[:,I]
      spcv = np.linalg.inv(kpcv).transpose()
   else:
      # user defined crystal structure
      spcv = primitiveCellVectors
      
   standardPrimitiveCellVectors = spcv.transpose()
   return standardPrimitiveCellVectors

def getStandardUnitCellVectors(crystalStructure,unitCellVectors,tol):
   ucv = unitCellVectors
   u1 = np.linalg.norm(ucv[:,0])
   u2 = np.linalg.norm(ucv[:,1])
   u3 = np.linalg.norm(ucv[:,2])
   if crystalStructure in ['Cubic','CUB','SimpleCubic','SC','cP','FaceCenteredCubic','FCC','cF','BodyCenteredCubic','BCC','cI']:
      sucv = ucv
      if np.linalg.det(sucv) < 0:
         sucv = sucv[:,[1,0,2]]
   elif crystalStructure in ['Tetragonal','TET','tP','BodyCenteredTetragonal','BCT','tI']:
      if abs(1-u1/u2) <= tol:
         sucv = ucv
      elif abs(1-u2/u3) <= tol:
         sucv = ucv[:,[1,2,0]]
      elif abs(1-u3/u1) <= tol:
         sucv = ucv[:,[2,0,1]]
      if np.linalg.det(sucv) < 0:
         sucv = sucv[:,[1,0,2]]
   elif crystalStructure in ['Orthorhombic','ORC','oP','FaceCenteredOrthorhombic','ORCF','oF','BodyCenteredOrthorhombic','ORCI','oI']:
      I = np.argsort([u1, u2, u3])
      sucv = ucv[:,I]
      if np.linalg.det(sucv) < 0:
         sucv[:,2] = -sucv[:,2]
   elif crystalStructure in ['CCenteredOrthorhombic','ORCC','oS']:
      if u1 < u2:
         sucv = ucv
      else:
         sucv = ucv[:,[1,0,2]]
      if np.linalg.det(sucv) < 0:
         sucv[:,2] = -sucv[:,2]
   standardUnitCellVectors = sucv
   return standardUnitCellVectors

###########
# testing #
###########

def bra2avec(bra):
   #BRA2AVEC Summary of this function goes here
   #   Detailed explanation goes here
   # rot = rand(3); [rot,~]=qr(rot,0)
   pi = np.pi
   rot = np.array([[-0.612200019964379, 0.598061781627136, 0.517236155844301],
   [-0.784920533612324, -0.380695311210428, -0.488846433899380],
   [-0.095450989881504, -0.705261076041611,  0.702492649891278]])
   noteye = np.logical_not(np.eye(3)).astype(np.float)
   if bra == 'CUB':
      avec = np.matmul(np.eye(3), rot) # 'CUB'
   elif bra == 'FCC':
      avec = np.matmul(noteye, rot) # 'FCC'
   elif bra == 'BCC':
      avec = np.matmul((noteye-np.eye(3)), rot) # 'BCC'
   elif bra == 'RHL1':
      avec = np.array([[np.cos(0/180*pi)*np.sin(30/180*pi), np.sin(0/180*pi)*np.sin(30/180*pi), np.cos(30/180*pi)],
         [np.cos(120/180*pi)*np.sin(30/180*pi), np.sin(120/180*pi)*np.sin(30/180*pi), np.cos(30/180*pi)],         
         [np.cos(240/180*pi)*np.sin(30/180*pi), np.sin(240/180*pi)*np.sin(30/180*pi), np.cos(30/180*pi)]])
      avec = np.matmul(avec, rot) # 'RHL1'
   elif bra == 'RHL2':
      avec = np.array([[np.cos(0/180*pi)*np.sin(60/180*pi), np.sin(0/180*pi)*np.sin(60/180*pi), np.cos(60/180*pi)],
         [np.cos(120/180*pi)*np.sin(60/180*pi), np.sin(120/180*pi)*np.sin(60/180*pi), np.cos(60/180*pi)],
         [np.cos(240/180*pi)*np.sin(60/180*pi), np.sin(240/180*pi)*np.sin(60/180*pi), np.cos(60/180*pi)]])
      avec = np.matmul(avec, rot) # 'RHL2'
   elif bra == 'HEX1':
      avec = np.array([[np.cos(0/180*pi), np.sin(0/180*pi), 0],
         [np.cos(120/180*pi), np.sin(120/180*pi), 0],
         [0, 0, 2]])
      avec = np.matmul(avec, rot) # 'HEX'
   elif bra == 'HEX2':
      avec = np.array([[np.cos(0/180*pi), np.sin(0/180*pi), 0],
         [np.cos(60/180*pi), np.sin(60/180*pi), 0],
         [0, 0, 2]])
      avec = np.matmul(avec, rot) # 'HEX'
   elif bra == 'TET1':
      avec = np.matmul(np.diag([1,1,2]), rot) # 'TET'
   elif bra == 'TET2':
      avec = np.matmul(np.diag([2,1,2]), rot) # 'TET'
   elif bra == 'BCT1':
      avec = np.matmul((noteye-np.eye(3)) * np.array([2,2,1]), rot) # 'BCT1'
   elif bra == 'BCT2':
      avec = np.matmul((noteye-np.eye(3)) * np.array([1,1,2]), rot) # 'BCT2'
   elif bra == 'ORC':
      avec = np.matmul(np.diag([1,2,3]), rot) # 'ORC'
   elif bra == 'ORCF1':
      avec = np.matmul(noteye * np.array([1,2,3]), rot) # 'ORCF1'
   elif bra == 'ORCF2':
      avec = np.matmul(noteye * np.array([1,1.1,1.2]), rot) # 'ORCF2'
   elif bra == 'ORCF3':
      avec = np.matmul(noteye * np.array([1,np.sqrt(4/3),2]), rot) # 'ORCF3'
   elif bra == 'ORCI':
      avec = np.matmul((noteye-np.eye(3)) * np.array([1,2,3]), rot) # 'ORCI'
   elif bra == 'ORCC':
      avec = np.array([[1, -1, 0], [1, 1, 0], [0, 0, 1]])
      avec = np.matmul(avec * np.array([1,2,3]), rot) # 'ORCC'
   elif bra == 'MCL':
      alpha = pi/5
      avec = np.array([[1, 0, 0], [0, 1, 0], [0, np.cos(alpha), np.sin(alpha)]])
      avec = np.matmul(avec * np.array([[1,2,3]]).transpose(), rot) # 'MCL'
   elif bra == 'MCLC1':
      alpha = pi/2*0.9
      avec = np.matmul([[1, 2, 0], [-1, 2, 0], [0, 5*np.cos(alpha), 5*np.sin(alpha)]], rot) # 'MCLC1'
   elif bra == 'MCLC2':
      alpha = pi/2*0.1
      avec = np.matmul([[1, 1, 0], [-1, 1, 0], [0, 5*np.cos(alpha), 5*np.sin(alpha)]], rot) # 'MCLC2'
   elif bra == 'MCLC3':
      avec = np.matmul([[5.97357258101562, 3.45227203834125, 0.00000000000000],
         [-5.97357258101562, 3.45227203834125, 0.00000000000000],
         [0.00000000000000, 2.27661214292493, 6.80081526960862]], rot) # 'MCLC3'
   elif bra == 'MCLC4':
      alpha = pi/2*0.15160094
      avec = np.matmul([[1, 2, 0], [-1, 2, 0], [0, 5*np.cos(alpha), 5*np.sin(alpha)]], rot) # 'MCLC4'
   elif bra == 'MCLC5':
      avec = np.matmul([[5.40176783775381, 4.52407218161051, 0.00000000000000],
         [-5.40176783775381, 4.52407218161051, 0.00000000000000],
         [0.00000000000000, 4.51116525071579, 4.56749911424154]], rot) # 'MCLC5'
   return avec

def print_all_bra2avec():
   perm = [[0,1,2],[0,2,1],[1,0,2],[1,2,0],[2,0,1],[2,1,0]]

   crystalStructure = 'CUB'; avec = bra2avec('CUB')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','X','M','R']

   crystalStructure = 'FCC'; avec = bra2avec('FCC')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','K','L','U','W','X']

   crystalStructure = 'BCC'; avec = bra2avec('BCC')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','H','P','N']

   crystalStructure = 'TET'; avec = bra2avec('TET1')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','A','M','R','X','Z']

   crystalStructure = 'TET'; avec = bra2avec('TET2')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','A','M','R','X','Z']

   crystalStructure = 'BCT'; avec = bra2avec('BCT1')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','M','N','P','X','Z','Z1']

   crystalStructure = 'BCT'; avec = bra2avec('BCT2')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','N','P','X','Z','S','S1','Y','Y1']

   crystalStructure = 'ORC'; avec = bra2avec('ORC')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','R','S','T','U','X','Y','Z']

   crystalStructure = 'ORCF'; avec = bra2avec('ORCF1')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','L','T','Y','Z','A','A1','X','X1']

   crystalStructure = 'ORCF'; avec = bra2avec('ORCF2')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','L','X','Y','Z','C','C1','D','D1','H','H1']

   crystalStructure = 'ORCF'; avec = bra2avec('ORCF3')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','L','T','Y','Z','A','A1','X','X1']

   crystalStructure = 'ORCI'; avec = bra2avec('ORCI')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','R','S','T','W','Z','L','L1','L2','X','X1','Y','Y1']

   crystalStructure = 'ORCC'; avec = bra2avec('ORCC')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','R','S','T','Y','Z','A','A1','X','X1']

   crystalStructure = 'HEX'; avec = bra2avec('HEX1')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','A','H','K','L','M']

   crystalStructure = 'HEX'; avec = bra2avec('HEX2')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','A','H','K','L','M']

   crystalStructure = 'RHL'; avec = bra2avec('RHL1')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','F','L','L1','Z','B','B1','P','P1','P2','Q','X']

   crystalStructure = 'RHL'; avec = bra2avec('RHL2')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','F','L','Z','P','P1','Q','Q1']

   crystalStructure = 'MCL'; avec = bra2avec('MCL')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','A','C','D','D1','E','X','Y','Y1','Z','H','H1','H2','M','M1','M2']

   crystalStructure = 'MCLC'; avec = bra2avec('MCLC1')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','N','N1','L','M','Y','Y1','Z','F','F1','F2','F3','I','I1','X','X1','X2']

   crystalStructure = 'MCLC'; avec = bra2avec('MCLC3')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','I','M','N','N1','X','Z','F','F1','F2','H','H1','H2','Y','Y1','Y2','Y3']

   crystalStructure = 'MCLC'; avec = bra2avec('MCLC5')
   for i in range(0, len(perm)):
      avecp = avec[perm[i],:]
      print(get3DCrystalStructure(avecp))
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecp.flatten(order='F').tolist()))
      avecstd = getStandardPrimitiveCellVectors(crystalStructure, avecp)
      print("%f %f %f %f %f %f %f %f %f " % tuple(avecstd.flatten(order='F').tolist()))
   kSymbols = ['G','L','M','N','N1','X','Z','F','F1','F2','H','H1','H2','I','I1','Y','Y1','Y2','Y3']




# -*- coding: utf-8 -*-
"""
Created on 2020-06-25

@author: Vincent Michaud-Rioux
"""

import numpy as np
import os
import shutil
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict

class Cmd:
   """
   command class.
   
   Attributes
   ----------
   path : string
      Path to the rescu installation.

   mpi : string
      MPI-launcher command.

   Methods
   -------
   bs, dos, scf
      Perform the eponym calculations.
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):

      # required status
      reqDict = {
         "mpi" : False,
         "stdout": False,
         "path": False,
      }
      # default values
      defDict = {
         "mpi" : "mpiexec -n 1",
         "stdout": None,
         "path": None,
      }    

      init_from_dict(self, reqDict, defDict, inpDict)

   def asdict(self):
      return utils_asdict(self)

   def bs(self, file):
      command = self.get_bs_cmd()
      command(file)

   def dos(self, file):
      command = self.get_dos_cmd()
      command(file)

   def scf(self, file):
      command = self.get_scf_cmd()
      command(file)
      
   def find_rescu(self):
      if self.path is None or len(self.path) == 0:
         fpath = shutil.which("rescu_scf")
      else:
         fpath = self.path+"/rescu_scf"
      found = os.path.isfile(fpath) and os.access(fpath, os.X_OK)
      if (not found):
         raise NameError("Cannot find rescu binary "+fpath+
         ". You may remedy the situation calling Rescu.set_cmd({'path':'/path/to/rescu'}).")

   def get_bs_cmd(self):
      return self.get_cmd("bs")

   def get_dos_cmd(self):
      return self.get_cmd("dos")

   def get_scf_cmd(self):
      return self.get_cmd("scf")

   def get_cmd(self, name):
      self.find_rescu()
      command = self.mpi+" "
      if self.path is None or len(self.path) == 0:
         command = command+" rescu_"+name
      else:
         command = command+self.path+"/rescu_"+name
      if self.stdout is None:
         return lambda file: os.system(command+" -i "+file)
      else:
         return lambda file: os.system(command+" -i "+file+" >> "+self.stdout)




# -*- coding: utf-8 -*-
"""
Created on 2020-05-12

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict

class Spin:
   """
   spin class.
   
   Attributes
   ----------
   ispin : integer
      Spin treatment level (1 : degenerate, 2 : collinear, 4 : non-collinear).

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):
      # required status
      reqDict = {
         "ispin": False,
      }
      # default values
      defDict = {
         "ispin": 1,
      }    

      init_from_dict(self, reqDict, defDict, inpDict)

      if not self.ispin in [1]: 
      # if not self.ispin in [1, 2, 4]: 
         raise NameError('Invalid spin.ispin value: '+self.ispin+'.')

   def asdict(self):
      return utils_asdict(self)

   def get_spin_num(self): 
      return 1


# -*- coding: utf-8 -*-
"""
Created on 2020-05-11

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict

class Eig:
   """
   eig class.
   
   Attributes
   ----------
   algo : string
      Algorithm.
   jobz : string
      If jobz = 'v', eigenvectors are calculated; only eigenvalues are returned otherwise.
   uplo : string
      If uplo = 'u', the upper triangular part of the matrices are referenced; if uplo = 'l', the upper triangular part of the matrices are referenced.
   range : string
      If range = 'a', all eigenvalues are computed; if range = 'i', the eigenvalues with index in the interval (il, iu) are computed; if range = 'v', the eigenvalues with index in the interval (vl, vu) are computed.
   ibtype : integer
      Eigenvalue problem type; if ibtype = 1, solve A*x = (lambda)*B*x; if ibtype = 2, solve A*B*x = (lambda)*x; if ibtype = 3, solve B*A*x = (lambda)*x.
   il : integer
      Lower index bound.
   iu : integer
      Upper index bound.
   lwork : integer
      Workspace size.
   reduAlgo : integer
      Switch between triangular inverse (1) and ScaLAPCK's sygst (2). 
   vl : float
      Lower value bound.
   vu : float
      Upper value bound.
   abstol : float
      The absolute error tolerance for the eigenvalues.
   orfac : float
      Specifies which eigenvectors should be reorthogonalized.
   trgtband : 1D array
      Target bands (must satisfy convergence criteria).
   inclband : 1D array
      Included bands (do not necessarily satisfy convergence criteria).

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):
      # required status
      reqDict = {
         "algo": False,
         # "jobz": False,
         # "uplo": False,
         # "range": False,
         # "ibtype": False,
         "reduAlgo": False,
         # "il": False,
         # "iu": False,
         "lwork": False,
         "orfac": False,
         "abstol": False,
         "trgtband": False,
         "inclband": False,
      }    
      # default values
      defDict = {
         "algo": "x",
         # "jobz": "n", 
         # "uplo": "u", 
         # "range": "a",
         # "ibtype": 1., 
         "reduAlgo": 2,
         # "il": 0, 
         # "iu": 0, 
         "lwork": -1,
         "orfac": 1.0e-6,
         "abstol": 1.0e-15,
         "trgtband": None,
         "inclband": None,
      }    
      
      init_from_dict(self, reqDict, defDict, inpDict)

   def asdict(self):
      return utils_asdict(self)


# -*- coding: utf-8 -*-
"""
Created on 2020-05-11

@author: Vincent Michaud-Rioux
"""

import numpy as np
import scipy.io as sio
from rescupy.species import Species
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict
from rescupy.utils import get_chemical_symbols

class Atoms:
   """
   atoms class.
   
   Attributes
   ----------
   coords : 2D array    
      Atomic coordinates (fractional).
   species : species-array     
      Object containing species related parameters.
   speciesPtr : 1D array  
      Species index of each atom (atom <==> species correspondance).
   ionCharge : float
      Total ionic charge.
   valCharge : float
      Total valence charge.

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict):
      # required status
      reqDict = {
         "coords": True,
         "speciesPtr": True,
         "valCharge": False,
         # "species": True,
      }
      # default values
      defDict = {
         "coords": None,
         "speciesPtr": None,
         "valCharge": None,
         # "species": {},
      }

      nspc = len(inpDict['species'])
      self.species = [Species(inpDict['species'][i]) for i in range(nspc)]
      
      init_from_dict(self, reqDict, defDict, inpDict)

      self.coords = np.array(self.coords).reshape((-1,3))
      self.speciesPtr = np.array(self.speciesPtr)
      self.ionCharge = sum(self.species[i].psp.Z for i in self.speciesPtr - 1)
      if self.valCharge is None:
         self.valCharge = self.ionCharge
      
      self.finalize_init()

   def asdict(self):
      return utils_asdict(self)

   def finalize_init(self):
      self.numspc = len(self.species)

   def get_positions(self, cell):
      return np.matmul(self.coords, cell.avec)

   def get_symbols(self, standard=False):
      if standard:
         symbols = [label_to_symbol(self.species[idx].label) for idx in self.speciesPtr - 1]
      else:
         symbols = [self.species[idx].label for idx in self.speciesPtr - 1]
      return symbols

def label_to_symbol(label):
   """Convert a label to an atomic species

   Parameters
   ----------
   label : str
      Should be an atomic species plus a tag. (e.g. H1, H_surf).

   Returns
   -------
   symbol : str
      The best matching species from the periodic table.

   Raises
   ------
   KeyError
      Couldn't find an appropriate species.

   """

   chemical_symbols = get_chemical_symbols()
   # two character species
   if len(label) >= 2:
      test_symbol = label[0].upper() + label[1].lower()
      if test_symbol in chemical_symbols:
         return test_symbol
   # one character species
   test_symbol = label[0].upper()
   if test_symbol in chemical_symbols:
      return test_symbol
   else:
      raise KeyError('Could not parse species from label {0}.'
                     ''.format(label))



# -*- coding: utf-8 -*-
"""
Created on 2020-05-11

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict

class Basis:
   """
   Basis class.
   
   Attributes
   ----------
   type : string
      Basis type (e.g. nao, pw).
   sprsthrs : float
      Sparsity threshold.

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):
      # required status
      reqDict = {
         "type"    : False,
         "sprsthrs": False,
      }    
      # default values
      defDict = {
         "type"    : 'nao',
         "sprsthrs": 0.15,
      }    
      
      init_from_dict(self, reqDict, defDict, inpDict)

   def asdict(self):
      return utils_asdict(self)


# -*- coding: utf-8 -*-
"""
Created on 2020-05-11

@author: Vincent Michaud-Rioux
"""

import copy
import numpy as np
from rescupy.orb import KbOrb, RadFunc
from rescupy.aobasis import VnlBasis
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict
from rescupy.utils import load_dcal

class Pspot:
   """
   pspot class.
   
   Attributes
   ----------
   symbol : string
      element symbol
   path : string
      pseudopotential path
   Z : integer
      pseudoion charge
   vlc : radFunc
      local pseudopotential
   rpc : radFunc
      partial core charge
   vnl : kbOrb-array
      KB-orbitals
   vso : kbOrb-array
      SO-orbitals

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict):
      # required status
      reqDict = {
         "path"  : True ,
         "symbol": False,
         "Z"     : False,
      }
      # default values
      defDict = {
         "path"  : None,
         "symbol": None,
         "Z"     : None,
      }
      
      init_from_dict(self, reqDict, defDict, inpDict)

      filename = self.path
      self.vlc = RadFunc(filename, 'Vlocal')
      self.rpc = RadFunc(filename, 'Rpc')
      self.vnl = VnlBasis({"symbol":self.symbol, "path":self.path}, 'Vnl')
      self.vso = copy.deepcopy(self.vnl); self.vso.orb = self.vso.orb[0:1]; self.vso.numorb = 1
      # self.vso = VnlBasis({"symbol":self.symbol, "path":self.path}, 'VnlSO')
      data, fmt = load_dcal(filename)
      for snam, inam in zip(['Z','symbol'],['N','symbol']):
         try:   
            if fmt == 'mat':
               tmp = data['atom'][0][0][inam][0][0]
            else:
               tmp = data['atom'][inam][0:]
               if inam == 'symbol':
                  tmp = ''.join(map(chr,tmp.flatten()))
            if isinstance(tmp, np.ndarray): tmp = float(tmp)
            if isinstance(tmp, np.str): tmp = str(tmp)
            setattr(self, snam, tmp)  
         except ValueError:
            raise NameError("Couldn't find parameter 'atom."+inam+" in file "+filename+".")

   def asdict(self):
      return utils_asdict(self)

         


# -*- coding: utf-8 -*-
"""
Created on 2020-05-12

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict

class Xc:
   """
   xc class.
   
   Attributes
   ----------
   funcid : 1D array
      Functional id (LibXC).

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):

      if "names" in inpDict.keys() and not "funcid" in inpDict.keys():
         inpDict["funcid"] = [name2id(name.lower()) for name in inpDict["names"]]

      # required status
      reqDict = {
         "funcid": False,
      }
      # default values
      defDict = {
         "funcid": [1, 12],
      }    

      init_from_dict(self, reqDict, defDict, inpDict)

   def asdict(self):
      return utils_asdict(self)

def name2id(name):
   idDict = {
      'xc_lda_x' :                         1,  # Exchange                                                              */
      'xc_lda_c_wigner' :                  2,  # Wigner parametrization                                                */
      'xc_lda_c_rpa' :                     3,  # Random Phase Approximation                                            */
      'xc_lda_c_hl' :                      4,  # Hedin & Lundqvist                                                     */
      'xc_lda_c_gl' :                      5,  # Gunnarson & Lundqvist                                                 */
      'xc_lda_c_xalpha' :                  6,  # Slater Xalpha                                                         */
      'xc_lda_c_vwn' :                     7,  # Vosko, Wilk, & Nusair (5)                                             */
      'xc_lda_c_vwn_rpa' :                 8,  # Vosko, Wilk, & Nusair (RPA)                                           */
      'xc_lda_c_pz' :                      9,  # Perdew & Zunger                                                       */
      'xc_lda_c_pz_mod' :                 10,  # Perdew & Zunger (Modified)                                            */
      'xc_lda_c_ob_pz' :                  11,  # Ortiz & Ballone (PZ)                                                  */
      'xc_lda_c_pw' :                     12,  # Perdew & Wang                                                         */
      'xc_lda_c_pw_mod' :                 13,  # Perdew & Wang (Modified)                                              */
      'xc_lda_c_ob_pw' :                  14,  # Ortiz & Ballone (PW)                                                  */
      'xc_lda_c_2d_amgb' :                15,  # Attaccalite et al                                                     */
      'xc_lda_c_2d_prm' :                 16,  # Pittalis, Rasanen & Marques correlation in 2D                         */
      'xc_lda_c_vbh' :                    17,  # von Barth & Hedin                                                     */
      'xc_lda_c_1d_csc' :                 18,  # Casula, Sorella, and Senatore 1D correlation                          */
      'xc_lda_x_2d' :                     19,  # Exchange in 2D                                                        */
      'xc_lda_xc_teter93' :               20,  # Teter 93 parametrization                                              */
      'xc_lda_x_1d' :                     21,  # Exchange in 1D                                                        */
      'xc_lda_c_ml1' :                    22,  # Modified LSD (version 1) of Proynov and Salahub                       */
      'xc_lda_c_ml2' :                    23,  # Modified LSD (version 2) of Proynov and Salahub                       */
      'xc_lda_c_gombas' :                 24,  # Gombas parametrization                                                */
      'xc_lda_c_pw_rpa' :                 25,  # Perdew & Wang fit of the RPA                                          */
      'xc_lda_c_1d_loos' :                26,  # P-F Loos correlation LDA                                              */
      'xc_lda_c_rc04' :                   27,  # Ragot-Cortona                                                         */
      'xc_lda_c_vwn_1' :                  28,  # Vosko, Wilk, & Nusair (1)                                             */
      'xc_lda_c_vwn_2' :                  29,  # Vosko, Wilk, & Nusair (2)                                             */
      'xc_lda_c_vwn_3' :                  30,  # Vosko, Wilk, & Nusair (3)                                             */
      'xc_lda_c_vwn_4' :                  31,  # Vosko, Wilk, & Nusair (4)                                             */
      'xc_lda_xc_zlp' :                   43,  # Zhao, Levy & Parr, Eq. (20)                                           */
      'xc_lda_k_tf' :                     50,  # Thomas-Fermi kinetic energy functional                                */
      'xc_lda_k_lp' :                     51,  # Lee and Parr Gaussian ansatz                                          */
      'xc_lda_xc_ksdt' :                 259,  # Karasiev et al. parametrization                                       */
      'xc_lda_c_chachiyo' :              287,  # Chachiyo simple 2 parameter correlation                               */
      'xc_lda_c_lp96' :                  289,  # Liu-Parr correlation                                                  */
      'xc_lda_x_rel' :                   532,  # Relativistic exchange                                                 */
      'xc_lda_xc_1d_ehwlrg_1' :          536,  # LDA constructed from slab-like systems of 1 electron                  */
      'xc_lda_xc_1d_ehwlrg_2' :          537,  # LDA constructed from slab-like systems of 2 electrons                 */
      'xc_lda_xc_1d_ehwlrg_3' :          538,  # LDA constructed from slab-like systems of 3 electrons                 */
      'xc_lda_x_erf' :                   546,  # Attenuated exchange LDA (erf)                                         */
      'xc_lda_xc_lp_a' :                 547,  # Lee-Parr reparametrization B                                          */
      'xc_lda_xc_lp_b' :                 548,  # Lee-Parr reparametrization B                                          */
      'xc_lda_x_rae' :                   549,  # Rae self-energy corrected exchange                                    */
      'xc_lda_k_zlp' :                   550,  # kinetic energy version of ZLP                                         */
      'xc_lda_c_mcweeny' :               551,  # McWeeny 76                                                            */
      'xc_lda_c_br78' :                  552,  # Brual & Rothstein 78                                                  */
      'xc_lda_c_pk09' :                  554,  # Proynov and Kong 2009                                                 */
      'xc_lda_c_ow_lyp' :                573,  # Wigner with corresponding LYP parameters                              */
      'xc_lda_c_ow' :                    574,  # Optimized Wigner                                                      */
      'xc_lda_xc_gdsmfb' :               577,  # Groth et al. parametrization                                          */
      'xc_lda_c_gk72' :                  578,  # Gordon and Kim 1972                                                   */
      'xc_lda_c_karasiev' :              579,  # Karasiev reparameterization of Chachiyo                               */
      'xc_lda_k_lp96' :                  580,  # Liu-Parr kinetic                                                      */
      'xc_gga_x_gam' :                    32,  # GAM functional from Minnesota                                         */
      'xc_gga_c_gam' :                    33,  # GAM functional from Minnesota                                         */
      'xc_gga_x_hcth_a' :                 34,  # HCTH-A                                                                */
      'xc_gga_x_ev93' :                   35,  # Engel and Vosko                                                       */
      'xc_gga_x_bcgp' :                   38,  # Burke, Cancio, Gould, and Pittalis                                    */
      'xc_gga_c_bcgp' :                   39,  # Burke, Cancio, Gould, and Pittalis                                    */
      'xc_gga_x_lambda_oc2_n' :           40,  # lambda_OC2(N) version of PBE                                          */
      'xc_gga_x_b86_r' :                  41,  # Revised Becke 86 Xalpha,beta,gamma (with mod. grad. correction)       */
      'xc_gga_x_lambda_ch_n' :            44,  # lambda_CH(N) version of PBE                                           */
      'xc_gga_x_lambda_lo_n' :            45,  # lambda_LO(N) version of PBE                                           */
      'xc_gga_x_hjs_b88_v2' :             46,  # HJS screened exchange corrected B88 version                           */
      'xc_gga_c_q2d' :                    47,  # Chiodo et al                                                          */
      'xc_gga_x_q2d' :                    48,  # Chiodo et al                                                          */
      'xc_gga_x_pbe_mol' :                49,  # Del Campo, Gazquez, Trickey and Vela (PBE-like)                       */
      'xc_gga_k_tfvw' :                   52,  # Thomas-Fermi plus von Weiszaecker correction                          */
      'xc_gga_k_revapbeint' :             53,  # interpolated version of REVAPBE                                       */
      'xc_gga_k_apbeint' :                54,  # interpolated version of APBE                                          */
      'xc_gga_k_revapbe' :                55,  # revised APBE                                                          */
      'xc_gga_x_ak13' :                   56,  # Armiento & Kuemmel 2013                                               */
      'xc_gga_k_meyer' :                  57,  # Meyer,  Wang, and Young                                               */
      'xc_gga_x_lv_rpw86' :               58,  # Berland and Hyldgaard                                                 */
      'xc_gga_x_pbe_tca' :                59,  # PBE revised by Tognetti et al                                         */
      'xc_gga_x_pbeint' :                 60,  # PBE for hybrid interfaces                                             */
      'xc_gga_c_zpbeint' :                61,  # spin-dependent gradient correction to PBEint                          */
      'xc_gga_c_pbeint' :                 62,  # PBE for hybrid interfaces                                             */
      'xc_gga_c_zpbesol' :                63,  # spin-dependent gradient correction to PBEsol                          */
      'xc_gga_xc_opbe_d' :                65,  # oPBE_D functional of Goerigk and Grimme                               */
      'xc_gga_xc_opwlyp_d' :              66,  # oPWLYP-D functional of Goerigk and Grimme                             */
      'xc_gga_xc_oblyp_d' :               67,  # oBLYP-D functional of Goerigk and Grimme                              */
      'xc_gga_x_vmt84_ge' :               68,  # VMT{8,4} with constraint satisfaction with mu = mu_GE                 */
      'xc_gga_x_vmt84_pbe' :              69,  # VMT{8,4} with constraint satisfaction with mu = mu_PBE                */
      'xc_gga_x_vmt_ge' :                 70,  # Vela, Medel, and Trickey with mu = mu_GE                              */
      'xc_gga_x_vmt_pbe' :                71,  # Vela, Medel, and Trickey with mu = mu_PBE                             */
      'xc_gga_c_n12_sx' :                 79,  # N12-SX functional from Minnesota                                      */
      'xc_gga_c_n12' :                    80,  # N12 functional from Minnesota                                         */
      'xc_gga_x_n12' :                    82,  # N12 functional from Minnesota                                         */
      'xc_gga_c_regtpss' :                83,  # Regularized TPSS correlation (ex-VPBE)                                */
      'xc_gga_c_op_xalpha' :              84,  # one-parameter progressive functional (XALPHA version)                 */
      'xc_gga_c_op_g96' :                 85,  # one-parameter progressive functional (G96 version)                    */
      'xc_gga_c_op_pbe' :                 86,  # one-parameter progressive functional (PBE version)                    */
      'xc_gga_c_op_b88' :                 87,  # one-parameter progressive functional (B88 version)                    */
      'xc_gga_c_ft97' :                   88,  # Filatov & Thiel correlation                                           */
      'xc_gga_c_spbe' :                   89,  # PBE correlation to be used with the SSB exchange                      */
      'xc_gga_x_ssb_sw' :                 90,  # Swart, Sola and Bickelhaupt correction to PBE                         */
      'xc_gga_x_ssb' :                    91,  # Swart, Sola and Bickelhaupt                                           */
      'xc_gga_x_ssb_d' :                  92,  # Swart, Sola and Bickelhaupt dispersion                                */
      'xc_gga_xc_hcth_407p' :             93,  # HCTH/407+                                                             */
      'xc_gga_xc_hcth_p76' :              94,  # HCTH p=7/6                                                            */
      'xc_gga_xc_hcth_p14' :              95,  # HCTH p=1/4                                                            */
      'xc_gga_xc_b97_gga1' :              96,  # Becke 97 GGA-1                                                        */
      'xc_gga_c_hcth_a' :                 97,  # HCTH-A                                                                */
      'xc_gga_x_bpccac' :                 98,  # BPCCAC (GRAC for the energy)                                          */
      'xc_gga_c_revtca' :                 99,  # Tognetti, Cortona, Adamo (revised)                                    */
      'xc_gga_c_tca' :                   100,  # Tognetti, Cortona, Adamo                                              */
      'xc_gga_x_pbe' :                   101,  # Perdew, Burke & Ernzerhof exchange                                    */
      'xc_gga_x_pbe_r' :                 102,  # Perdew, Burke & Ernzerhof exchange (revised)                          */
      'xc_gga_x_b86' :                   103,  # Becke 86 Xalpha,beta,gamma                                            */
      'xc_gga_x_herman' :                104,  # Herman et al original GGA                                             */
      'xc_gga_x_b86_mgc' :               105,  # Becke 86 Xalpha,beta,gamma (with mod. grad. correction)               */
      'xc_gga_x_b88' :                   106,  # Becke 88                                                              */
      'xc_gga_x_g96' :                   107,  # Gill 96                                                               */
      'xc_gga_x_pw86' :                  108,  # Perdew & Wang 86                                                      */
      'xc_gga_x_pw91' :                  109,  # Perdew & Wang 91                                                      */
      'xc_gga_x_optx' :                  110,  # Handy & Cohen OPTX 01                                                 */
      'xc_gga_x_dk87_r1' :               111,  # dePristo & Kress 87 (version R1)                                      */
      'xc_gga_x_dk87_r2' :               112,  # dePristo & Kress 87 (version R2)                                      */
      'xc_gga_x_lg93' :                  113,  # Lacks & Gordon 93                                                     */
      'xc_gga_x_ft97_a' :                114,  # Filatov & Thiel 97 (version A)                                        */
      'xc_gga_x_ft97_b' :                115,  # Filatov & Thiel 97 (version B)                                        */
      'xc_gga_x_pbe_sol' :               116,  # Perdew, Burke & Ernzerhof exchange (solids)                           */
      'xc_gga_x_rpbe' :                  117,  # Hammer, Hansen & Norskov (PBE-like)                                   */
      'xc_gga_x_wc' :                    118,  # Wu & Cohen                                                            */
      'xc_gga_x_mpw91' :                 119,  # Modified form of PW91 by Adamo & Barone                               */
      'xc_gga_x_am05' :                  120,  # Armiento & Mattsson 05 exchange                                       */
      'xc_gga_x_pbea' :                  121,  # Madsen (PBE-like)                                                     */
      'xc_gga_x_mpbe' :                  122,  # Adamo & Barone modification to PBE                                    */
      'xc_gga_x_xpbe' :                  123,  # xPBE reparametrization by Xu & Goddard                                */
      'xc_gga_x_2d_b86_mgc' :            124,  # Becke 86 MGC for 2D systems                                           */
      'xc_gga_x_bayesian' :              125,  # Bayesian best fit for the enhancement factor                          */
      'xc_gga_x_pbe_jsjr' :              126,  # JSJR reparametrization by Pedroza, Silva & Capelle                    */
      'xc_gga_x_2d_b88' :                127,  # Becke 88 in 2D                                                        */
      'xc_gga_x_2d_b86' :                128,  # Becke 86 Xalpha,beta,gamma                                            */
      'xc_gga_x_2d_pbe' :                129,  # Perdew, Burke & Ernzerhof exchange in 2D                              */
      'xc_gga_c_pbe' :                   130,  # Perdew, Burke & Ernzerhof correlation                                 */
      'xc_gga_c_lyp' :                   131,  # Lee, Yang & Parr                                                      */
      'xc_gga_c_p86' :                   132,  # Perdew 86                                                             */
      'xc_gga_c_pbe_sol' :               133,  # Perdew, Burke & Ernzerhof correlation SOL                             */
      'xc_gga_c_pw91' :                  134,  # Perdew & Wang 91                                                      */
      'xc_gga_c_am05' :                  135,  # Armiento & Mattsson 05 correlation                                    */
      'xc_gga_c_xpbe' :                  136,  # xPBE reparametrization by Xu & Goddard                                */
      'xc_gga_c_lm' :                    137,  # Langreth and Mehl correlation                                         */
      'xc_gga_c_pbe_jrgx' :              138,  # JRGX reparametrization by Pedroza, Silva & Capelle                    */
      'xc_gga_x_optb88_vdw' :            139,  # Becke 88 reoptimized to be used with vdW functional of Dion et al     */
      'xc_gga_x_pbek1_vdw' :             140,  # PBE reparametrization for vdW                                         */
      'xc_gga_x_optpbe_vdw' :            141,  # PBE reparametrization for vdW                                         */
      'xc_gga_x_rge2' :                  142,  # Regularized PBE                                                       */
      'xc_gga_c_rge2' :                  143,  # Regularized PBE                                                       */
      'xc_gga_x_rpw86' :                 144,  # refitted Perdew & Wang 86                                             */
      'xc_gga_x_kt1' :                   145,  # Exchange part of Keal and Tozer version 1                             */
      'xc_gga_xc_kt2' :                  146,  # Keal and Tozer version 2                                              */
      'xc_gga_c_wl' :                    147,  # Wilson & Levy                                                         */
      'xc_gga_c_wi' :                    148,  # Wilson & Ivanov                                                       */
      'xc_gga_x_mb88' :                  149,  # Modified Becke 88 for proton transfer                                 */
      'xc_gga_x_sogga' :                 150,  # Second-order generalized gradient approximation                       */
      'xc_gga_x_sogga11' :               151,  # Second-order generalized gradient approximation 2011                  */
      'xc_gga_c_sogga11' :               152,  # Second-order generalized gradient approximation 2011                  */
      'xc_gga_c_wi0' :                   153,  # Wilson & Ivanov initial version                                       */
      'xc_gga_xc_th1' :                  154,  # Tozer and Handy v. 1                                                  */
      'xc_gga_xc_th2' :                  155,  # Tozer and Handy v. 2                                                  */
      'xc_gga_xc_th3' :                  156,  # Tozer and Handy v. 3                                                  */
      'xc_gga_xc_th4' :                  157,  # Tozer and Handy v. 4                                                  */
      'xc_gga_x_c09x' :                  158,  # C09x to be used with the VdW of Rutgers-Chalmers                      */
      'xc_gga_c_sogga11_x' :             159,  # To be used with HYB_GGA_X_SOGGA11_X                                   */
      'xc_gga_x_lb' :                    160,  # van Leeuwen & Baerends                                                */
      'xc_gga_xc_hcth_93' :              161,  # HCTH functional fitted to  93 molecules                               */
      'xc_gga_xc_hcth_120' :             162,  # HCTH functional fitted to 120 molecules                               */
      'xc_gga_xc_hcth_147' :             163,  # HCTH functional fitted to 147 molecules                               */
      'xc_gga_xc_hcth_407' :             164,  # HCTH functional fitted to 407 molecules                               */
      'xc_gga_xc_edf1' :                 165,  # Empirical functionals from Adamson, Gill, and Pople                   */
      'xc_gga_xc_xlyp' :                 166,  # XLYP functional                                                       */
      'xc_gga_xc_kt1' :                  167,  # Keal and Tozer version 1                                              */
      'xc_gga_xc_b97_d' :                170,  # Grimme functional to be used with C6 vdW term                         */
      'xc_gga_xc_pbe1w' :                173,  # Functionals fitted for water                                          */
      'xc_gga_xc_mpwlyp1w' :             174,  # Functionals fitted for water                                          */
      'xc_gga_xc_pbelyp1w' :             175,  # Functionals fitted for water                                          */
      'xc_gga_x_lbm' :                   182,  # van Leeuwen & Baerends modified                                       */
      'xc_gga_x_ol2' :                   183,  # Exchange form based on Ou-Yang and Levy v.2                           */
      'xc_gga_x_apbe' :                  184,  # mu fixed from the semiclassical neutral atom                          */
      'xc_gga_k_apbe' :                  185,  # mu fixed from the semiclassical neutral atom                          */
      'xc_gga_c_apbe' :                  186,  # mu fixed from the semiclassical neutral atom                          */
      'xc_gga_k_tw1' :                   187,  # Tran and Wesolowski set 1 (Table II)                                  */
      'xc_gga_k_tw2' :                   188,  # Tran and Wesolowski set 2 (Table II)                                  */
      'xc_gga_k_tw3' :                   189,  # Tran and Wesolowski set 3 (Table II)                                  */
      'xc_gga_k_tw4' :                   190,  # Tran and Wesolowski set 4 (Table II)                                  */
      'xc_gga_x_htbs' :                  191,  # Haas, Tran, Blaha, and Schwarz                                        */
      'xc_gga_x_airy' :                  192,  # Constantin et al based on the Airy gas                                */
      'xc_gga_x_lag' :                   193,  # Local Airy Gas                                                        */
      'xc_gga_xc_mohlyp' :               194,  # Functional for organometallic chemistry                               */
      'xc_gga_xc_mohlyp2' :              195,  # Functional for barrier heights                                        */
      'xc_gga_xc_th_fl' :                196,  # Tozer and Handy v. FL                                                 */
      'xc_gga_xc_th_fc' :                197,  # Tozer and Handy v. FC                                                 */
      'xc_gga_xc_th_fcfo' :              198,  # Tozer and Handy v. FCFO                                               */
      'xc_gga_xc_th_fco' :               199,  # Tozer and Handy v. FCO                                                */
      'xc_gga_c_optc' :                  200,  # Optimized correlation functional of Cohen and Handy                   */
      'xc_gga_c_pbeloc' :                246,  # Semilocal dynamical correlation                                       */
      'xc_gga_xc_vv10' :                 255,  # Vydrov and Van Voorhis                                                */
      'xc_gga_c_pbefe' :                 258,  # PBE for formation energies                                            */
      'xc_gga_c_op_pw91' :               262,  # one-parameter progressive functional (PW91 version)                   */
      'xc_gga_x_pbefe' :                 265,  # PBE for formation energies                                            */
      'xc_gga_x_cap' :                   270,  # Correct Asymptotic Potential                                          */
      'xc_gga_x_eb88' :                  271,  # Non-empirical (excogitated) B88 functional of Becke and Elliott       */
      'xc_gga_c_pbe_mol' :               272,  # Del Campo, Gazquez, Trickey and Vela (PBE-like)                       */
      'xc_gga_k_absp3' :                 277,  # gamma-TFvW form by Acharya et al [g = 1 - 1.513/N^0.35]               */
      'xc_gga_k_absp4' :                 278,  # gamma-TFvW form by Acharya et al [g = l = 1/(1 + 1.332/N^(1/3))]      */
      'xc_gga_c_bmk' :                   280,  # Boese-Martin for kinetics                                             */
      'xc_gga_c_tau_hcth' :              281,  # correlation part of tau-hcth                                          */
      'xc_gga_c_hyb_tau_hcth' :          283,  # correlation part of hyb_tau-hcth                                      */
      'xc_gga_x_beefvdw' :               285,  # BEEF-vdW exchange                                                     */
      'xc_gga_xc_beefvdw' :              286,  # BEEF-vdW exchange-correlation                                         */
      'xc_gga_x_pbetrans' :              291,  # Gradient-based interpolation between PBE and revPBE                   */
      'xc_gga_x_chachiyo' :              298,  # Chachiyo exchange                                                     */
      'xc_gga_k_vw' :                    500,  # von Weiszaecker functional                                            */
      'xc_gga_k_ge2' :                   501,  # Second-order gradient expansion (l = 1/9)                             */
      'xc_gga_k_golden' :                502,  # TF-lambda-vW form by Golden (l = 13/45)                               */
      'xc_gga_k_yt65' :                  503,  # TF-lambda-vW form by Yonei and Tomishima (l = 1/5)                    */
      'xc_gga_k_baltin' :                504,  # TF-lambda-vW form by Baltin (l = 5/9)                                 */
      'xc_gga_k_lieb' :                  505,  # TF-lambda-vW form by Lieb (l = 0.185909191)                           */
      'xc_gga_k_absp1' :                 506,  # gamma-TFvW form by Acharya et al [g = 1 - 1.412/N^(1/3)]              */
      'xc_gga_k_absp2' :                 507,  # gamma-TFvW form by Acharya et al [g = 1 - 1.332/N^(1/3)]              */
      'xc_gga_k_gr' :                    508,  # gamma-TFvW form by Gazquez and Robles                                 */
      'xc_gga_k_ludena' :                509,  # gamma-TFvW form by Ludena                                             */
      'xc_gga_k_gp85' :                  510,  # gamma-TFvW form by Ghosh and Parr                                     */
      'xc_gga_k_pearson' :               511,  # Pearson                                                               */
      'xc_gga_k_ol1' :                   512,  # Ou-Yang and Levy v.1                                                  */
      'xc_gga_k_ol2' :                   513,  # Ou-Yang and Levy v.2                                                  */
      'xc_gga_k_fr_b88' :                514,  # Fuentealba & Reyes (B88 version)                                      */
      'xc_gga_k_fr_pw86' :               515,  # Fuentealba & Reyes (PW86 version)                                     */
      'xc_gga_k_dk' :                    516,  # DePristo and Kress                                                    */
      'xc_gga_k_perdew' :                517,  # Perdew                                                                */
      'xc_gga_k_vsk' :                   518,  # Vitos, Skriver, and Kollar                                            */
      'xc_gga_k_vjks' :                  519,  # Vitos, Johansson, Kollar, and Skriver                                 */
      'xc_gga_k_ernzerhof' :             520,  # Ernzerhof                                                             */
      'xc_gga_k_lc94' :                  521,  # Lembarki & Chermette                                                  */
      'xc_gga_k_llp' :                   522,  # Lee, Lee & Parr                                                       */
      'xc_gga_k_thakkar' :               523,  # Thakkar 1992                                                          */
      'xc_gga_x_wpbeh' :                 524,  # short-range version of the PBE                                        */
      'xc_gga_x_hjs_pbe' :               525,  # HJS screened exchange PBE version                                     */
      'xc_gga_x_hjs_pbe_sol' :           526,  # HJS screened exchange PBE_SOL version                                 */
      'xc_gga_x_hjs_b88' :               527,  # HJS screened exchange B88 version                                     */
      'xc_gga_x_hjs_b97x' :              528,  # HJS screened exchange B97x version                                    */
      'xc_gga_x_ityh' :                  529,  # short-range recipe for exchange GGA functionals                       */
      'xc_gga_x_sfat' :                  530,  # short-range recipe for exchange GGA functionals                       */
      'xc_gga_x_sg4' :                   533,  # Semiclassical GGA at fourth order                                     */
      'xc_gga_c_sg4' :                   534,  # Semiclassical GGA at fourth order                                     */
      'xc_gga_x_gg99' :                  535,  # Gilbert and Gill 1999                                                 */
      'xc_gga_x_pbepow' :                539,  # PBE power                                                             */
      'xc_gga_x_kgg99' :                 544,  # Gilbert and Gill 1999 (mixed)                                         */
      'xc_gga_xc_hle16' :                545,  # high local exchange 2016                                              */
      'xc_gga_c_scan_e0' :               553,  # GGA component of SCAN                                                 */
      'xc_gga_c_gapc' :                  555,  # GapC                                                                  */
      'xc_gga_c_gaploc' :                556,  # Gaploc                                                                */
      'xc_gga_c_zvpbeint' :              557,  # another spin-dependent correction to PBEint                           */
      'xc_gga_c_zvpbesol' :              558,  # another spin-dependent correction to PBEsol                           */
      'xc_gga_c_tm_lyp' :                559,  # Takkar and McCarthy reparametrization                                 */
      'xc_gga_c_tm_pbe' :                560,  # Thakkar and McCarthy reparametrization                                */
      'xc_gga_c_w94' :                   561,  # Wilson 94 (Eq. 25)                                                    */
      'xc_gga_c_cs1' :                   565,  # A dynamical correlation functional                                    */
      'xc_gga_x_b88m' :                  570,  # Becke 88 reoptimized to be used with mgga_c_tau1                      */
      'xc_gga_k_pbe3' :                  595,  # Three parameter PBE-like expansion                                    */
      'xc_gga_k_pbe4' :                  596,  # Four  parameter PBE-like expansion                                    */
      'xc_gga_k_exp4' :                  597,  # Intermediate form between PBE3 and PBE4                               */
      'xc_hyb_gga_x_n12_sx' :             81,  # N12-SX functional from Minnesota                                      */
      'xc_hyb_gga_xc_b97_1p' :           266,  # version of B97 by Cohen and Handy                                     */
      'xc_hyb_gga_xc_pbe_mol0' :         273,  # PBEmol0                                                               */
      'xc_hyb_gga_xc_pbe_sol0' :         274,  # PBEsol0                                                               */
      'xc_hyb_gga_xc_pbeb0' :            275,  # PBEbeta0                                                              */
      'xc_hyb_gga_xc_pbe_molb0' :        276,  # PBEmolbeta0                                                           */
      'xc_hyb_gga_xc_pbe50' :            290,  # PBE0 with 50% exx                                                     */
      'xc_hyb_gga_xc_b3pw91' :           401,  # The original (ACM) hybrid of Becke                                    */
      'xc_hyb_gga_xc_b3lyp' :            402,  # The (in)famous B3LYP                                                  */
      'xc_hyb_gga_xc_b3p86' :            403,  # Perdew 86 hybrid similar to B3PW91                                    */
      'xc_hyb_gga_xc_o3lyp' :            404,  # hybrid using the optx functional                                      */
      'xc_hyb_gga_xc_mpw1k' :            405,  # mixture of mPW91 and PW91 optimized for kinetics                      */
      'xc_hyb_gga_xc_pbeh' :             406,  # aka PBE0 or PBE1PBE                                                   */
      'xc_hyb_gga_xc_b97' :              407,  # Becke 97                                                              */
      'xc_hyb_gga_xc_b97_1' :            408,  # Becke 97-1                                                            */
      'xc_hyb_gga_xc_b97_2' :            410,  # Becke 97-2                                                            */
      'xc_hyb_gga_xc_x3lyp' :            411,  # hybrid by Xu and Goddard                                              */
      'xc_hyb_gga_xc_b1wc' :             412,  # Becke 1-parameter mixture of WC and PBE                               */
      'xc_hyb_gga_xc_b97_k' :            413,  # Boese-Martin for Kinetics                                             */
      'xc_hyb_gga_xc_b97_3' :            414,  # Becke 97-3                                                            */
      'xc_hyb_gga_xc_mpw3pw' :           415,  # mixture with the mPW functional                                       */
      'xc_hyb_gga_xc_b1lyp' :            416,  # Becke 1-parameter mixture of B88 and LYP                              */
      'xc_hyb_gga_xc_b1pw91' :           417,  # Becke 1-parameter mixture of B88 and PW91                             */
      'xc_hyb_gga_xc_mpw1pw' :           418,  # Becke 1-parameter mixture of mPW91 and PW91                           */
      'xc_hyb_gga_xc_mpw3lyp' :          419,  # mixture of mPW and LYP                                                */
      'xc_hyb_gga_xc_sb98_1a' :          420,  # Schmider-Becke 98 parameterization 1a                                 */
      'xc_hyb_gga_xc_sb98_1b' :          421,  # Schmider-Becke 98 parameterization 1b                                 */
      'xc_hyb_gga_xc_sb98_1c' :          422,  # Schmider-Becke 98 parameterization 1c                                 */
      'xc_hyb_gga_xc_sb98_2a' :          423,  # Schmider-Becke 98 parameterization 2a                                 */
      'xc_hyb_gga_xc_sb98_2b' :          424,  # Schmider-Becke 98 parameterization 2b                                 */
      'xc_hyb_gga_xc_sb98_2c' :          425,  # Schmider-Becke 98 parameterization 2c                                 */
      'xc_hyb_gga_x_sogga11_x' :         426,  # Hybrid based on SOGGA11 form                                          */
      'xc_hyb_gga_xc_hse03' :            427,  # the 2003 version of the screened hybrid HSE                           */
      'xc_hyb_gga_xc_hse06' :            428,  # the 2006 version of the screened hybrid HSE                           */
      'xc_hyb_gga_xc_hjs_pbe' :          429,  # HJS hybrid screened exchange PBE version                              */
      'xc_hyb_gga_xc_hjs_pbe_sol' :      430,  # HJS hybrid screened exchange PBE_SOL version                          */
      'xc_hyb_gga_xc_hjs_b88' :          431,  # HJS hybrid screened exchange B88 version                              */
      'xc_hyb_gga_xc_hjs_b97x' :         432,  # HJS hybrid screened exchange B97x version                             */
      'xc_hyb_gga_xc_cam_b3lyp' :        433,  # CAM version of B3LYP                                                  */
      'xc_hyb_gga_xc_tuned_cam_b3lyp' :  434,  # CAM version of B3LYP tuned for excitations                            */
      'xc_hyb_gga_xc_bhandh' :           435,  # Becke half-and-half                                                   */
      'xc_hyb_gga_xc_bhandhlyp' :        436,  # Becke half-and-half with B88 exchange                                 */
      'xc_hyb_gga_xc_mb3lyp_rc04' :      437,  # B3LYP with RC04 LDA                                                   */
      'xc_hyb_gga_xc_mpwlyp1m' :         453,  # MPW with 1 par. for metals/LYP                                        */
      'xc_hyb_gga_xc_revb3lyp' :         454,  # Revised B3LYP                                                         */
      'xc_hyb_gga_xc_camy_blyp' :        455,  # BLYP with yukawa screening                                            */
      'xc_hyb_gga_xc_pbe0_13' :          456,  # PBE0-1/3                                                              */
      'xc_hyb_gga_xc_b3lyps' :           459,  # B3LYP* functional                                                     */
      'xc_hyb_gga_xc_wb97' :             463,  # Chai and Head-Gordon                                                  */
      'xc_hyb_gga_xc_wb97x' :            464,  # Chai and Head-Gordon                                                  */
      'xc_hyb_gga_xc_lrc_wpbeh' :        465,  # Long-range corrected functional by Rorhdanz et al                     */
      'xc_hyb_gga_xc_wb97x_v' :          466,  # Mardirossian and Head-Gordon                                          */
      'xc_hyb_gga_xc_lcy_pbe' :          467,  # PBE with yukawa screening                                             */
      'xc_hyb_gga_xc_lcy_blyp' :         468,  # BLYP with yukawa screening                                            */
      'xc_hyb_gga_xc_lc_vv10' :          469,  # Vydrov and Van Voorhis                                                */
      'xc_hyb_gga_xc_camy_b3lyp' :       470,  # B3LYP with Yukawa screening                                           */
      'xc_hyb_gga_xc_wb97x_d' :          471,  # Chai and Head-Gordon                                                  */
      'xc_hyb_gga_xc_hpbeint' :          472,  # hPBEint                                                               */
      'xc_hyb_gga_xc_lrc_wpbe' :         473,  # Long-range corrected functional by Rorhdanz et al                     */
      'xc_hyb_gga_xc_b3lyp5' :           475,  # B3LYP with VWN functional 5 instead of RPA                            */
      'xc_hyb_gga_xc_edf2' :             476,  # Empirical functional from Lin, George and Gill                        */
      'xc_hyb_gga_xc_cap0' :             477,  # Correct Asymptotic Potential hybrid                                   */
      'xc_hyb_gga_xc_lc_wpbe' :          478,  # Long-range corrected functional by Vydrov and Scuseria                */
      'xc_hyb_gga_xc_hse12' :            479,  # HSE12 by Moussa, Schultz and Chelikowsky                              */
      'xc_hyb_gga_xc_hse12s' :           480,  # Short-range HSE12 by Moussa, Schultz, and Chelikowsky                 */
      'xc_hyb_gga_xc_hse_sol' :          481,  # HSEsol functional by Schimka, Harl, and Kresse                        */
      'xc_hyb_gga_xc_cam_qtp_01' :       482,  # CAM-QTP(01): CAM-B3LYP retuned using ionization potentials of water   */
      'xc_hyb_gga_xc_mpw1lyp' :          483,  # Becke 1-parameter mixture of mPW91 and LYP                            */
      'xc_hyb_gga_xc_mpw1pbe' :          484,  # Becke 1-parameter mixture of mPW91 and PBE                            */
      'xc_hyb_gga_xc_kmlyp' :            485,  # Kang-Musgrave hybrid                                                  */
      'xc_hyb_gga_xc_b5050lyp' :         572,  # Like B3LYP but more exact exchange                                    */
      'xc_mgga_c_dldf' :                  37,  # Dispersionless Density Functional                                     */
      'xc_mgga_xc_zlp' :                  42,  # Zhao, Levy & Parr, Eq. (21)                                           */
      'xc_mgga_xc_otpss_d' :              64,  # oTPSS_D functional of Goerigk and Grimme                              */
      'xc_mgga_c_cs' :                    72,  # Colle and Salvetti                                                    */
      'xc_mgga_c_mn12_sx' :               73,  # MN12-SX correlation functional from Minnesota                         */
      'xc_mgga_c_mn12_l' :                74,  # MN12-L correlation functional from Minnesota                          */
      'xc_mgga_c_m11_l' :                 75,  # M11-L correlation functional from Minnesota                           */
      'xc_mgga_c_m11' :                   76,  # M11 correlation functional from Minnesota                             */
      'xc_mgga_c_m08_so' :                77,  # M08-SO correlation functional from Minnesota                          */
      'xc_mgga_c_m08_hx' :                78,  # M08-HX correlation functional from Minnesota                          */
      'xc_mgga_x_lta' :                  201,  # Local tau approximation of Ernzerhof & Scuseria                       */
      'xc_mgga_x_tpss' :                 202,  # Tao, Perdew, Staroverov & Scuseria exchange                           */
      'xc_mgga_x_m06_l' :                203,  # M06-L exchange functional from Minnesota                              */
      'xc_mgga_x_gvt4' :                 204,  # GVT4 from Van Voorhis and Scuseria                                    */
      'xc_mgga_x_tau_hcth' :             205,  # tau-HCTH from Boese and Handy                                         */
      'xc_mgga_x_br89' :                 206,  # Becke-Roussel 89                                                      */
      'xc_mgga_x_bj06' :                 207,  # Becke & Johnson correction to Becke-Roussel 89                        */
      'xc_mgga_x_tb09' :                 208,  # Tran & Blaha correction to Becke & Johnson                            */
      'xc_mgga_x_rpp09' :                209,  # Rasanen, Pittalis, and Proetto correction to Becke & Johnson          */
      'xc_mgga_x_2d_prhg07' :            210,  # Pittalis, Rasanen, Helbig, Gross Exchange Functional                  */
      'xc_mgga_x_2d_prhg07_prp10' :      211,  # PRGH07 with PRP10 correction                                          */
      'xc_mgga_x_revtpss' :              212,  # revised Tao, Perdew, Staroverov & Scuseria exchange                   */
      'xc_mgga_x_pkzb' :                 213,  # Perdew, Kurth, Zupan, and Blaha                                       */
      'xc_mgga_x_ms0' :                  221,  # MS exchange of Sun, Xiao, and Ruzsinszky                              */
      'xc_mgga_x_ms1' :                  222,  # MS1 exchange of Sun, et al                                            */
      'xc_mgga_x_ms2' :                  223,  # MS2 exchange of Sun, et al                                            */
      'xc_mgga_x_m11_l' :                226,  # M11-L exchange functional from Minnesota                              */
      'xc_mgga_x_mn12_l' :               227,  # MN12-L exchange functional from Minnesota                             */
      'xc_mgga_xc_cc06' :                229,  # Cancio and Chou 2006                                                  */
      'xc_mgga_x_mk00' :                 230,  # Exchange for accurate virtual orbital energies                        */
      'xc_mgga_c_tpss' :                 231,  # Tao, Perdew, Staroverov & Scuseria correlation                        */
      'xc_mgga_c_vsxc' :                 232,  # VSxc from Van Voorhis and Scuseria (correlation part)                 */
      'xc_mgga_c_m06_l' :                233,  # M06-L correlation functional from Minnesota                           */
      'xc_mgga_c_m06_hf' :               234,  # M06-HF correlation functional from Minnesota                          */
      'xc_mgga_c_m06' :                  235,  # M06 correlation functional from Minnesota                             */
      'xc_mgga_c_m06_2x' :               236,  # M06-2X correlation functional from Minnesota                          */
      'xc_mgga_c_m05' :                  237,  # M05 correlation functional from Minnesota                             */
      'xc_mgga_c_m05_2x' :               238,  # M05-2X correlation functional from Minnesota                          */
      'xc_mgga_c_pkzb' :                 239,  # Perdew, Kurth, Zupan, and Blaha                                       */
      'xc_mgga_c_bc95' :                 240,  # Becke correlation 95                                                  */
      'xc_mgga_c_revtpss' :              241,  # revised TPSS correlation                                              */
      'xc_mgga_xc_tpsslyp1w' :           242,  # Functionals fitted for water                                          */
      'xc_mgga_x_mk00b' :                243,  # Exchange for accurate virtual orbital energies (v. B)                 */
      'xc_mgga_x_bloc' :                 244,  # functional with balanced localization                                 */
      'xc_mgga_x_modtpss' :              245,  # Modified Tao, Perdew, Staroverov & Scuseria exchange                  */
      'xc_mgga_c_tpssloc' :              247,  # Semilocal dynamical correlation                                       */
      'xc_mgga_x_mbeef' :                249,  # mBEEF exchange                                                        */
      'xc_mgga_x_mbeefvdw' :             250,  # mBEEF-vdW exchange                                                    */
      'xc_mgga_xc_b97m_v' :              254,  # Mardirossian and Head-Gordon                                          */
      'xc_mgga_x_mvs' :                  257,  # MVS exchange of Sun, Perdew, and Ruzsinszky                           */
      'xc_mgga_x_mn15_l' :               260,  # MN15-L exhange functional from Minnesota                              */
      'xc_mgga_c_mn15_l' :               261,  # MN15-L correlation functional from Minnesota                          */
      'xc_mgga_x_scan' :                 263,  # SCAN exchange of Sun, Ruzsinszky, and Perdew                          */
      'xc_mgga_c_scan' :                 267,  # SCAN correlation                                                      */
      'xc_mgga_c_mn15' :                 269,  # MN15 correlation functional from Minnesota                            */
      'xc_mgga_x_b00' :                  284,  # Becke 2000                                                            */
      'xc_mgga_xc_hle17' :               288,  # high local exchange 2017                                              */
      'xc_mgga_c_scan_rvv10' :           292,  # SCAN correlation + rVV10 correlation                                  */
      'xc_mgga_x_revm06_l' :             293,  # revised M06-L exchange functional from Minnesota                      */
      'xc_mgga_c_revm06_l' :             294,  # Revised M06-L correlation functional from Minnesota                   */
      'xc_mgga_x_tm' :                   540,  # Tao and Mo 2016                                                       */
      'xc_mgga_x_vt84' :                 541,  # meta-GGA version of VT{8,4} GGA                                       */
      'xc_mgga_x_sa_tpss' :              542,  # TPSS with correct surface asymptotics                                 */
      'xc_mgga_k_pc07' :                 543,  # Perdew and Constantin 2007                                            */
      'xc_mgga_c_kcis' :                 562,  # Krieger, Chen, Iafrate, and Savin                                     */
      'xc_mgga_xc_lp90' :                564,  # Lee & Parr, Eq. (56)                                                  */
      'xc_mgga_c_b88' :                  571,  # Meta-GGA correlation by Becke                                         */
      'xc_mgga_x_gx' :                   575,  # GX functional of Loos                                                 */
      'xc_mgga_x_pbe_gx' :               576,  # PBE-GX functional of Loos                                             */
      'xc_mgga_x_revscan' :              581,  # revised SCAN                                                          */
      'xc_mgga_c_revscan' :              582,  # revised SCAN correlation                                              */
      'xc_mgga_c_scan_vv10' :            584,  # SCAN correlation +  VV10 correlation                                  */
      'xc_mgga_c_revscan_vv10' :         585,  # revised SCAN correlation                                              */
      'xc_mgga_x_br89_explicit' :        586,  # Becke-Roussel 89 with an explicit inversion of x(y)                   */
      'xc_hyb_mgga_x_dldf' :              36,  # Dispersionless Density Functional                                     */
      'xc_hyb_mgga_x_ms2h' :             224,  # MS2 hybrid exchange of Sun, et al                                     */
      'xc_hyb_mgga_x_mn12_sx' :          248,  # MN12-SX hybrid exchange functional from Minnesota                     */
      'xc_hyb_mgga_x_scan0' :            264,  # SCAN hybrid exchange                                                  */
      'xc_hyb_mgga_x_mn15' :             268,  # MN15 hybrid exchange functional from Minnesota                        */
      'xc_hyb_mgga_x_bmk' :              279,  # Boese-Martin for kinetics                                             */
      'xc_hyb_mgga_x_tau_hcth' :         282,  # Hybrid version of tau-HCTH                                            */
      'xc_hyb_mgga_x_m08_hx' :           295,  # M08-HX exchange functional from Minnesota                             */
      'xc_hyb_mgga_x_m08_so' :           296,  # M08-SO exchange functional from Minnesota                             */
      'xc_hyb_mgga_x_m11' :              297,  # M11 hybrid exchange functional from Minnesota                         */
      'xc_hyb_mgga_x_m05' :              438,  # M05 hybrid exchange functional from Minnesota                         */
      'xc_hyb_mgga_x_m05_2x' :           439,  # M05-2X hybrid exchange functional from Minnesota                      */
      'xc_hyb_mgga_xc_b88b95' :          440,  # Mixture of B88 with BC95 (B1B95)                                      */
      'xc_hyb_mgga_xc_b86b95' :          441,  # Mixture of B86 with BC95                                              */
      'xc_hyb_mgga_xc_pw86b95' :         442,  # Mixture of PW86 with BC95                                             */
      'xc_hyb_mgga_xc_bb1k' :            443,  # Mixture of B88 with BC95 from Zhao and Truhlar                        */
      'xc_hyb_mgga_x_m06_hf' :           444,  # M06-HF hybrid exchange functional from Minnesota                      */
      'xc_hyb_mgga_xc_mpw1b95' :         445,  # Mixture of mPW91 with BC95 from Zhao and Truhlar                      */
      'xc_hyb_mgga_xc_mpwb1k' :          446,  # Mixture of mPW91 with BC95 for kinetics                               */
      'xc_hyb_mgga_xc_x1b95' :           447,  # Mixture of X with BC95                                                */
      'xc_hyb_mgga_xc_xb1k' :            448,  # Mixture of X with BC95 for kinetics                                   */
      'xc_hyb_mgga_x_m06' :              449,  # M06 hybrid exchange functional from Minnesota                         */
      'xc_hyb_mgga_x_m06_2x' :           450,  # M06-2X hybrid exchange functional from Minnesota                      */
      'xc_hyb_mgga_xc_pw6b95' :          451,  # Mixture of PW91 with BC95 from Zhao and Truhlar                       */
      'xc_hyb_mgga_xc_pwb6k' :           452,  # Mixture of PW91 with BC95 from Zhao and Truhlar for kinetics          */
      'xc_hyb_mgga_xc_tpssh' :           457,  # TPSS hybrid                                                           */
      'xc_hyb_mgga_xc_revtpssh' :        458,  # revTPSS hybrid                                                        */
      'xc_hyb_mgga_x_mvsh' :             474,  # MVSh hybrid                                                           */
      'xc_hyb_mgga_xc_wb97m_v' :         531,  # Mardirossian and Head-Gordon                                          */
      'xc_hyb_mgga_xc_b0kcis' :          563,  # Hybrid based on KCIS                                                  */
      'xc_hyb_mgga_xc_mpw1kcis' :        566,  # Modified Perdew-Wang + KCIS hybrid                                    */
      'xc_hyb_mgga_xc_mpwkcis1k' :       567,  # Modified Perdew-Wang + KCIS hybrid with more exact exchange           */
      'xc_hyb_mgga_xc_pbe1kcis' :        568,  # Perdew-Burke-Ernzerhof + KCIS hybrid                                  */
      'xc_hyb_mgga_xc_tpss1kcis' :       569,  # TPSS hybrid with KCIS correlation                                     */
      'xc_hyb_mgga_x_revscan0' :         583,  # revised SCAN hybrid exchange                                          */
      'xc_hyb_mgga_xc_b98' :             598,  # Becke 98                                                              */
   }
   return idDict[name]



# -*- coding: utf-8 -*-
"""
Created on 2020-05-11

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.atoms   import Atoms
from rescupy.cell    import Cell
from rescupy.dos     import Dos
from rescupy.energy  import Energy
from rescupy.kmesh   import Kmesh
from rescupy.pop     import Pop
from rescupy.species import Species
from rescupy.spin    import Spin
from rescupy.xc      import Xc
from rescupy.utils   import asdict    as utils_asdict
from rescupy.utils import init_from_dict
from rescupy.utils   import set_units as utils_set_units
import scipy.io as sio
from matplotlib import pyplot as plt

class AtomSys:
   """
   atomSys class.
   
   Attributes
   ----------
   cell : cell object           
      Object containing cell related parameters.
   atoms : atoms object
      Object describing the atom collection (coordinates, species, etc.)
   kmesh : kmesh-object          
      Object containing kmesh related parameters.
   pop : pop-object 
      Object containing population (occupancy) related parameters.

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict):
      # required status
      reqDict = {
         "atoms"  : True,
         "cell"   : True,
         "dos"    : False,
         "energy" : False,
         "kmesh"  : False,
         "pop"    : False,
         "spin"   : False,
         "xc"     : False,
      }
      # default values
      defDict = {
         "atoms"  : {},
         "cell"   : {},
         "dos"    : {},
         "energy" : {},
         "kmesh"  : {},
         "pop"    : {},
         "spin"   : {},
         "xc"     : {},
      }

      init_from_dict(self, reqDict, defDict, inpDict)

      self.finalize_init()

      self.check_init()

   def asdict(self):
      return utils_asdict(self)

   def finalize_init(self):
      self.kmesh.set_bvec(self.cell)
      if self.kmesh.type == "line":
         self.kmesh.set_bz_lines(self.cell)

   def check_init(self):
      if (self.pop.type in ['tm'] and self.kmesh.grid == None):
         raise NameError('kmesh.grid must be provided if pop.type == "tm"')

      if (self.pop.type in ['tm'] and np.product(self.kmesh.grid) < 8):
         raise NameError('kmesh.grid must be at least 2 in every dimension if pop.type == "tm"')

   def plot_bs(self):
      evals = self.get_evals()
      evals = evals[:,:,0].T 
      # extract labels
      chpts = self.kmesh.checkPoints
      kpts = np.array(self.kmesh.kpts).reshape((-1,3))
      kpts = [kpts[k,:] for k in chpts]
      labels = self.kmesh.kpt_2_label(self.cell, kpts)
      # remove doublons (e.g. 'X' following 'X')
      for i in range(0, len(chpts)-1):
         if chpts[i] + 1 == chpts[i+1] and labels[i] == labels[i+1]:
            j = chpts[i]
            evals = np.delete(evals, j, axis=0)
            del chpts[i]
            del labels[i]
            for k in range(i, len(chpts)): chpts[k] = chpts[k] - 1
         if i + 2 >= len(chpts): break
      ef = np.ones(evals.shape[0]) * self.energy.efermi
      # plot
      fig = plt.figure()
      plt.plot(evals, "-b", label="_nolegend_")
      plt.plot(ef, "--k", label="Efermi")
      plt.legend(loc="upper right")
      plt.xlabel("BZ ("+self.kmesh.bvecUnits+")")
      plt.ylabel("Energy ("+self.energy.evalsUnits+")")
      plt.xticks(chpts, labels)
      plt.grid(axis='x')
      plt.show()
      return fig

   def set_units(self, units):
      return utils_set_units(self, units)

   def get_evals(self):
      nkpt = self.kmesh.get_kpts_num()
      nspin = self.spin.get_spin_num()
      evals = self.energy.evals
      if isinstance(evals, list):
         evals = np.array(evals)
      if isinstance(evals, np.ndarray): 
         evals = evals.reshape((-1,nkpt,nspin),order="F")
      return evals

   def get_vbm(self):
      evals = self.get_evals()
      lvals = evals < self.energy.efermi
      if np.all(lvals == lvals[:,0:1,0:1]):
         vbm = np.max(evals[lvals])
      else:
         vbm = None
      return vbm



# Source:
# https://medium.com/@joel.barmettler/how-to-upload-your-python-package-to-pypi-65edc5fe9c56
# Example:
# pip install --user twine
# python setup.py sdist
# twine upload dist/*
from distutils.core import setup
setup(
   name = 'rescupy',       # How you named your package folder (MyLib)
   packages = ['rescupy'], # Chose the same as "name"
   version = '0.1.1',      # Start with a small number and increase it with every change you make
   license='Proprietary',  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
   description = 'RESCUPy is a Python interface for the Fortran-2008 version of RESCU.',   # Give a short description about your library
   author = 'Vincent Michaud-Rioux',               # Type in your name
   author_email = 'vincentm@nanoacademic.com',     # Type in your E-Mail
   url = 'https://github.com/vincentmr/rescupy',   # Provide either the link to your github or to your website
   download_url = 'https://github.com/vincentmr/rescupy/archive/v0.1.1.tar.gz',  # I explain this later on
   keywords = ['RESCU', 'RESCUF08', 'RESCUPy', 'DFT'],   # Keywords that define your package best
   install_requires=[                                    # I get to this in a second
      'h5py',
      'matplotlib',
      'numpy',
      'scipy',
         ],
   classifiers=[
      'Development Status :: 3 - Alpha',        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
      'Intended Audience :: Developers',        # Define that your audience are developers
      'Topic :: Software Development :: Build Tools',
      'License :: Other/Proprietary License',   # Again, pick a license
      'Programming Language :: Python :: 3',    # Specify which pyhton versions that you want to support
      'Programming Language :: Python :: 3.4',
      'Programming Language :: Python :: 3.5',
      'Programming Language :: Python :: 3.6',
  ],
)


# -*- coding: utf-8 -*-
"""
Created on 2020-06-16

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict
from rescupy.utils import set_units as utils_set_units

class Energy:
   """
   energy class.
   
   Attributes
   ----------
   esr : float
      Short-range energy.
   ebs : float
      Band structure energy.
   edh : float
      Delta Hartree energy.
   exc : float
      Exchange-correlation energy.
   evxc : float
      Exchange-correlation potential energy.
   etot : float
      Total energy.
   efermi : float
      Fermi energy.
   evals : 3D array
      Kohn-Sham energies.
   frc : 2D array
      Atomic forces.
   
   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict = {}):
      # required status
      reqDict = {
         "esr"         : False,
         "esrUnits"    : False,
         "ebs"         : False,
         "ebsUnits"    : False,
         "edh"         : False,
         "edhUnits"    : False,
         "exc"         : False,
         "excUnits"    : False,
         "evxc"        : False,
         "evxcUnits"   : False,
         "etot"        : False,
         "etotUnits"   : False,
         "efermi"      : False,
         "efermiUnits" : False,
         "evals"       : False,
         "evalsUnits"  : False,
         "efree"       : False, 
         "efreeUnits"  : False, 
         "entropy"     : False, 
         "entropyUnits": False, 
         "frc"         : False,
         "frcUnits"    : False,
         "frcReturn"   : False,
         "stress"      : False,
         "stressUnits" : False,
         "stressReturn": False,
      }
      # default values
      defDict = {
         "esr"         : None,
         "esrUnits"    : "ev",
         "ebs"         : None,
         "ebsUnits"    : "ev",
         "edh"         : None,
         "edhUnits"    : "ev",
         "exc"         : None,
         "excUnits"    : "ev",
         "evxc"        : None,
         "evxcUnits"   : "ev",
         "etot"        : None,
         "etotUnits"   : "ev",
         "efree"       : None,
         "efreeUnits"  : "ev",
         "entropy"     : None,
         "entropyUnits": "ev",
         "efermi"      : None,
         "efermiUnits" : "ev",
         "evals"       : None,
         "evalsUnits"  : "ev",
         "frc"         : None,
         "frcUnits"    : "ev/ang",
         "frcReturn"   : False,
         "stress"      : None,
         "stressUnits" : "ev/ang3",
         "stressReturn": False,
      }    

      init_from_dict(self, reqDict, defDict, inpDict)

      if isinstance(self.evals, list): 
         self.evals = np.array(self.evals)
         
      if isinstance(self.frc, list): 
         self.frc = np.array(self.frc).reshape((-1,3))
         
      if isinstance(self.stress, list): 
         self.stress = np.array(self.stress).reshape((-1,3))
         
   def asdict(self):
      return utils_asdict(self)

   def set_units(self, units):
      return utils_set_units(self, units)



# -*- coding: utf-8 -*-
"""
Created on 2020-05-11

@author: Vincent Michaud-Rioux
"""

import numpy as np
from rescupy.orb import AoOrb, KbOrb, RadFunc
from rescupy.utils import asdict as utils_asdict
from rescupy.utils import init_from_dict
from rescupy.utils import load_dcal

class NaoBasis:
   """
   naoBasis class.
   
   Attributes
   ----------
   symbol : string
      Element symbol.
   path : string
      Pseudopotential path.
   orb : aoOrb-array
      Atomic orbitals.

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict, varname):
      # default values
      defDict = {
         "path": None,
         "symbol": None,
      }
      # required status
      reqDict = {
         "path": True,
         "symbol": False,
      }
      
      init_from_dict(self, reqDict, defDict, inpDict)

      filename = self.path
      data, fmt = load_dcal(filename)
      if fmt == 'mat':
         norb = len(data[varname][0][0])
      else:
         norb = len(data[varname]["Parameter"])
      self.orb = [AoOrb(filename, varname, index=i) for i in range(norb)]
      for snam, inam in zip(['symbol'],['symbol']):
         try:   
            if fmt == 'mat':
               tmp = data['atom'][0][0][inam][0][0]
            else:
               tmp = data['atom'][inam][0:]
               if inam == 'symbol':
                  tmp = ''.join(map(chr,tmp.flatten()))
            if isinstance(tmp, np.ndarray): tmp = float(tmp)
            if isinstance(tmp, np.str): tmp = str(tmp)
            setattr(self, snam, tmp)  
         except ValueError:
            raise NameError("Couldn't find parameter 'atom."+inam+" in file "+filename+".")

      self.finalize_init()

   def asdict(self):
      return utils_asdict(self)

   def finalize_init(self):
      self.numorb = len(self.orb)

class VnlBasis:
   """
   vnlBasis class.
   
   Attributes
   ----------
   symbol : string
      Element symbol.
   path : string
      Pseudopotential path.
   orb : aoOrb-array
      Atomic orbitals.

   Methods
   -------
   
   Remarks
   ---

   """
   def __init__(self, inpDict, varname):
      # required status
      reqDict = {
         "path": True,
         "symbol": False,
      }
      # default values
      defDict = {
         "path": None,
         "symbol": None,
      }
      
      init_from_dict(self, reqDict, defDict, inpDict)

      filename = self.path
      data, fmt = load_dcal(filename)
      norb = 0
      if fmt == 'mat' and len(data[varname][0]) > 0:
         norb = len(data[varname][0][0])
      elif len(data[varname]["Parameter"]) > 0:
         # find length based on HDF5 object reference
         norb = len(data[varname]["Parameter"])
      if varname == 'VnlSO':
         rng = range(1,norb)
      else:
         rng = range(0,norb)
      self.orb = [KbOrb(filename, varname, index=i) for i in rng] # 1st entry is empty
      for snam, inam in zip(['symbol'],['symbol']):
         try:   
            if fmt == 'mat':
               tmp = data['atom'][0][0][inam][0][0]
            else:
               tmp = data['atom'][inam][0:]
               if inam == 'symbol':
                  tmp = ''.join(map(chr,tmp.flatten()))
            if isinstance(tmp, np.ndarray): tmp = float(tmp)
            if isinstance(tmp, np.str): tmp = str(tmp)
            setattr(self, snam, tmp)  
         except ValueError:
            raise NameError("Couldn't find parameter 'atom."+inam+" in file "+filename+".")

      self.finalize_init()

   def asdict(self):
      return utils_asdict(self)

   def finalize_init(self):
      self.numorb = len(self.orb)
